# Licensed under an MIT open source license - see LICENSE

import timeit
import os

from ._testdata import data

from .. import Dendrogram


def benchmark_compute():
    print("Data loaded. Starting dendrogram computations...")

    def test1():
        Dendrogram.compute(data, min_npix=4, min_value=1.4, min_delta=0.3)
        print("  Completed an iteration of test1.")

    def test2():
        Dendrogram.compute(data, min_npix=8, min_value=1.4)
        print("  Completed an iteration of test2.")

    num = 3

    t1 = timeit.timeit(test1, number=num) / num
    t2 = timeit.timeit(test2, number=num) / num

    print("test1 average over {num} computations: {result:.3} s".format(num=num, result=t1))
    print("test2 average over {num} computations: {result:.3} s".format(num=num, result=t2))

    print("Total average compute time: {0:.3}s".format((t1 + t2) / 2))


def benchmark_hdf5():
    print("\nGenerating complex dendrogram for HDF5 import/export...")
    d = Dendrogram.compute(data, min_npix=2, min_value=1.4, min_delta=0.01)
    print("Dendrogram generated. Testing import and export...")

    filename = '.astrodendro-hdf5-benchmark.hdf5'
    if os.path.exists(filename):
        os.remove(filename)
    num = 2

    def testHDF5():
        print('Exporting...')
        d.save_to(filename)
        print('Importing...')
        d2 = Dendrogram.load_from(filename)
        os.remove(filename)

    t = timeit.timeit(testHDF5, number=num) / num

    print("Total average export+import time: {0:.3}s".format(t))


if __name__ == '__main__':
    try:
        benchmark_compute()
        benchmark_hdf5()
    except KeyboardInterrupt:
        print("Cancelled.")



# Licensed under an MIT open source license - see LICENSE

import numpy as np


class DendrogramPlotter(object):

    """
    A class to plot a dendrogram object.
    """

    def __init__(self, dendrogram):
        # should we copy to ensure immutability?
        self.dendrogram = dendrogram
        self._cached_positions = None
        self.sort()

    def set_custom_positions(self, custom_position):
        """
        Manually set the positon on the structures for plotting.

        Parameters
        ----------
        custom_position : function
            This should be a function that takes a
            `~astrodendro.structure.Structure`returns the position of the
            leaves to use for plotting. If the dataset has more than one
            dimension, using this may cause lines to cross. If this is used,
            then ``sort_key`` and ``reverse`` are ignored.
        """
        self._cached_positions = {}
        for structure in self.dendrogram.all_structures:
            self._cached_positions[structure] = custom_position(structure)

    def sort(self, sort_key=None, reverse=False):
        """
        Sort the position of the leaves for plotting.

        Parameters
        ----------
        sort_key : function, optional
             This should be a function that takes a
             `~astrodendro.structure.Structure` and returns a scalar that is
             then used to sort the leaves. If not specified, the leaves are
             sorted according to their peak value.
        reverse : bool, optional
             Whether to reverse the sorting
        """

        if sort_key is None:
            sort_key = lambda s: s.get_peak(subtree=True)[1]

        sorted_trunk_structures = sorted(self.dendrogram.trunk, key=sort_key, reverse=reverse)

        positions = {}
        x = 0  # the first index for each trunk structure
        for structure in sorted_trunk_structures:

            # Get sorted leaves
            sorted_leaves = structure.sorted_leaves(subtree=True, reverse=reverse)

            # Loop over leaves and assign positions
            for leaf in sorted_leaves:
                positions[leaf] = x
                x += 1

            # Sort structures from the top-down
            sorted_structures = sorted(structure.descendants, key=lambda s: s.level, reverse=True) + [structure]

            # Loop through structures and assing position of branches as the mean
            # of the leaves
            for structure in sorted_structures:
                if not structure.is_leaf:
                    positions[structure] = np.mean([positions[child] for child in structure.children])

        self._cached_positions = positions

    def plot_tree(self, ax, structure=None, subtree=True, autoscale=True, **kwargs):
        """
        Plot the dendrogram tree or a substructure.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes` instance
            The Axes inside which to plot the dendrogram
        structure : int or `~astrodendro.structure.Structure`, optional
            If specified, only plot this structure. This can be either the
            structure object itself, or the ID (``idx``) of the structure.
        subtree : bool, optional
            If a structure is specified, by default the whole subtree will be
            plotted, but this can be disabled with this option.
        autoscale : bool, optional
            Whether to automatically adapt the window limits to the tree

        Notes
        -----
        Any additional keyword arguments are passed to
        `~matplotlib.collections.LineCollection` and can be used to control the
        appearance of the plot.
        """

        # Get the lines for the dendrogram
        lines = self.get_lines(structure=structure, **kwargs)

        # Add the lines to the axes
        ax.add_collection(lines)

        # Auto-scale axes (doesn't happen by default with ``add_collection``)
        if autoscale:
            ax.margins(0.05)
            ax.autoscale_view(True, True, True)

    def plot_contour(self, ax, structure=None, subtree=True, slice=None, **kwargs):
        """
        Plot a contour outlining all pixels in the dendrogram, or a specific.
        structure.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes` instance
            The Axes inside which to plot the dendrogram
        structure : int or `~astrodendro.structure.Structure`, optional
            If specified, only plot this structure. This can be either the
            structure object itself, or the ID (``idx``) of the structure.
        subtree : bool, optional
            If a structure is specified, by default the whole subtree will be
            plotted, but this can be disabled with this option.
        slice : int, optional
            If dealing with a 3-d cube, the slice at which to plot the contour.
            If not set, the slice containing the peak of the structure will be
            shown

        Notes
        -----
        Any additional keyword arguments are passed to
        `~matplotlib.axes.Axes.contour` and can be used to control the
        appearance of the plot.

        """
        if self.dendrogram.data.ndim not in [2, 3]:
            raise ValueError("plot_data can only be used with 2- or 3-dimensional data")

        if structure is None:
            mask = self.dendrogram.data > self.dendrogram.params['min_value']
        else:
            if type(structure) is int:
                structure = self.dendrogram[structure]
            mask = structure.get_mask(subtree=subtree)
            if self.dendrogram.data.ndim == 3:
                if slice is None:
                    peak_index = structure.get_peak(subtree=subtree)
                    slice = peak_index[0][0]
                mask = mask[slice, :, :]

        # fix a common mistake when trying to set the color of contours
        if 'color' in kwargs and 'colors' not in kwargs:
            kwargs['colors'] = kwargs['color']

        ax.contour(mask, levels=[0.5], **kwargs)

    def get_lines(self, structure=None, **kwargs):
        """
        Get a collection of lines to draw the dendrogram.

        Parameters
        ----------
        structure : `~astrodendro.structure.Structure`
            The structure to plot. If not set, the whole tree will be plotted.

        Returns
        -------
        lines : `astrodendro.plot.StructureCollection`
            The lines (sub-class of LineCollection) which can be directly used in Matplotlib

        Notes
        -----
        Any additional keyword arguments are passed to the
        `~matplotlib.collections.LineCollection` class.
        """

        if self._cached_positions is None:
            raise Exception("Leaves have not yet been sorted")

        if structure is None:
            structures = list(self.dendrogram.all_structures)
        else:
            if type(structure) is int:
                structure = self.dendrogram[structure]
            structures = structure.descendants + [structure]

        lines = []
        mapping = []
        for s in structures:
            x = self._cached_positions[s]
            bot = s.parent.height if s.parent is not None else s.vmin
            top = s.height
            lines.append(([x, bot], [x, top]))
            mapping.append(s)
            if s.is_branch:
                pc = [self._cached_positions[c] for c in s.children]
                lines.append(([min(pc), top], [max(pc), top]))
                mapping.append(s)

        from .structure_collection import StructureCollection
        sc = StructureCollection(lines, **kwargs)
        sc.structures = mapping
        return sc



# Licensed under an MIT open source license - see LICENSE

import os

import numpy as np

from .util import parse_dendrogram
from .handler import IOHandler

HDF5_SIGNATURE = b'\x89HDF\r\n\x1a\n'


def is_hdf5(filename, mode='r'):
    if mode == 'r' and os.path.exists(filename):
        fileobj = open(filename, 'rb')
        sig = fileobj.read(8)
        return sig == HDF5_SIGNATURE
    elif filename.lower().endswith(('.hdf5', '.h5')):
        return True
    else:
        return False


def dendro_export_hdf5(d, filename):
    """Export the dendrogram 'd' to the HDF5 file 'filename'"""
    import h5py
    f = h5py.File(filename, 'w')

    f.attrs['n_dim'] = d.n_dim

    f.create_dataset('newick', data=d.to_newick())

    ds = f.create_dataset('index_map', data=d.index_map, compression=True)
    ds.attrs['CLASS'] = 'IMAGE'
    ds.attrs['IMAGE_VERSION'] = '1.2'
    ds.attrs['IMAGE_MINMAXRANGE'] = [d.index_map.min(), d.index_map.max()]

    ds = f.create_dataset('data', data=d.data, compression=True)
    ds.attrs['CLASS'] = 'IMAGE'
    ds.attrs['IMAGE_VERSION'] = '1.2'
    ds.attrs['IMAGE_MINMAXRANGE'] = [d.data.min(), d.data.max()]

    f.close()


def dendro_import_hdf5(filename):
    """Import 'filename' and construct a dendrogram from it"""
    import h5py

    with h5py.File(filename, 'r') as h5f:
        newick = h5f['newick'].value
        data = h5f['data'].value
        index_map = h5f['index_map'].value

    return parse_dendrogram(newick, data, index_map)


HDF5Handler = IOHandler(identify=is_hdf5,
                        export_dendro=dendro_export_hdf5,
                        import_dendro=dendro_import_hdf5)



import numpy as np

from astropy import units as u
from astropy.constants import si


def quantity_sum(quantities):
    """
    In Astropy 0.3, np.sum will do the right thing for quantities, but in the mean time we need a workaround.
    """
    return np.sum(quantities.value) * quantities.unit


def compute_flux(input_quantities, output_unit, wavelength=None, spatial_scale=None,
                 velocity_scale=None, beam_major=None, beam_minor=None):
    """
    Given a set of flux values in arbitrary units, find the total flux in a
    specific set of units.

    Parameters
    ----------
    input_quantities : `~astropy.units.quantity.Quantity` instance
        A `~astropy.units.quantity.Quantity` instance containing an array of
        flux values to be summed.
    output_unit : `~astropy.units.core.Unit` instance
        The final unit to give the total flux in (should be equivalent to Jy)
    wavelength : `~astropy.units.quantity.Quantity` instance
        The wavelength of the data (required if converting e.g.
        ergs/cm^2/s/micron to Jy)
    spatial_scale : `~astropy.units.quantity.Quantity` instance
        The pixel scale of the data (should be an angle)
    velocity_scale : `~astropy.units.quantity.Quantity` instance
        The pixel scale of the data (should be a velocity)
    beam_major : `~astropy.units.quantity.Quantity` instance
        The beam major full width at half_maximum (FWHM)
    beam_minor : `~astropy.units.quantity.Quantity` instance
        The beam minor full width at half_maximum (FWHM)
    """

    # Start off by finding the total flux in Jy

    if input_quantities.unit.is_equivalent(u.Jy):  # Fnu

        # Simply sum up the values and convert to output unit
        total_flux = quantity_sum(input_quantities).to(u.Jy)

    elif input_quantities.unit.is_equivalent(u.erg / u.cm ** 2 / u.s / u.m):  # Flambda

        if wavelength is not None and not wavelength.unit.is_equivalent(u.m):
            raise ValueError("wavelength should be a physical length")

        # Find the frequency
        if wavelength is None:
            raise ValueError("wavelength is needed to convert from {0} to Jy".format(input_quantities.unit))

        # Find frequency
        nu = si.c / wavelength

        # Convert input quantity to Fnu in Jy
        q = (input_quantities * wavelength / nu).to(u.Jy)

        # Find total flux in Jy
        total_flux = quantity_sum(q)

    elif input_quantities.unit.is_equivalent(u.MJy / u.sr):  # surface brightness (Fnu)

        if spatial_scale is not None and not spatial_scale.unit.is_equivalent(u.degree):
            raise ValueError("spatial_scale should be an angle")

        if spatial_scale is None:
            raise ValueError("spatial_scale is needed to convert from {0} to Jy".format(input_quantities.unit))

        # Find the area of a pixel as a solid angle
        pixel_area = (spatial_scale ** 2)

        # Convert input quantity to Fnu in Jy
        q = (input_quantities * pixel_area).to(u.Jy)

        # Find total flux in Jy
        total_flux = quantity_sum(q)

    elif input_quantities.unit.is_equivalent(u.Jy / u.beam):

        if spatial_scale is not None and not spatial_scale.unit.is_equivalent(u.degree):
            raise ValueError("spatial_scale should be an angle")

        if spatial_scale is None:
            raise ValueError("spatial_scale is needed to convert from {0} to Jy".format(input_quantities.unit))

        if beam_major is not None and not beam_major.unit.is_equivalent(u.degree):
            raise ValueError("beam_major should be an angle")

        if beam_major is None:
            raise ValueError("beam_major is needed to convert from {0} to Jy".format(input_quantities.unit))

        if beam_minor is not None and not beam_minor.unit.is_equivalent(u.degree):
            raise ValueError("beam_minor should be an angle")

        if beam_minor is None:
            raise ValueError("beam_minor is needed to convert from {0} to Jy".format(input_quantities.unit))

        # Find the beam area
        beams_per_pixel = spatial_scale ** 2 / (beam_minor * beam_major * 1.1331) * u.beam

        # Convert input quantity to Fnu in Jy
        q = (input_quantities * beams_per_pixel).to(u.Jy)

        # Find total flux in Jy
        total_flux = quantity_sum(q)

    else:

        raise ValueError("Flux units {0} not yet supported".format(input_quantities.unit))

    if not output_unit.is_equivalent(u.Jy):
        raise ValueError("output_unit has to be equivalent to Jy")
    else:
        return total_flux.to(output_unit)



import os

import numpy as np
from astropy.io import fits

from astrodendro import Dendrogram

BENCHMARKS = {'2d1.fits': {'min_value': 2.5, 'min_npix': 20},
              '2d2.fits': {'min_value': 2.5},
              '2d3.fits': {'min_value': 2.5, 'min_delta': 1},
              '3d1.fits': {'min_value': 4.5, 'min_npix': 20},
              '3d2.fits': {'min_value': 4.5},
              '3d3.fits': {'min_value': 4.5, 'min_delta':1}}

def main():

    data = fits.getdata(os.path.join('benchmark_data', '2d.fits'))
    for outfile in '2d1.fits 2d2.fits 2d3.fits'.split():
        d = Dendrogram.compute(data, verbose=True, **BENCHMARKS[outfile])
        d.save_to(os.path.join('benchmark_data', outfile))

    data = fits.getdata(os.path.join('benchmark_data', '3d.fits'))
    for outfile in '3d1.fits 3d2.fits 3d3.fits'.split():
        d = Dendrogram.compute(data, verbose=True, **BENCHMARKS[outfile])
        d.save_to(os.path.join('benchmark_data', outfile))


if __name__ == "__main__":
    main()



# These tests ensure that the ``is_independent`` function is working correctly

import pytest
import numpy as np

from ..dendrogram import Dendrogram


class TestCustomMerge(object):

    def setup_class(self):
        self.data = np.array([0, 3.3, 5.5, 2.1, 1.0, 6.0, 4.4, 1.5, 4.1, 0.5])

    def test_reference(self):

        d = Dendrogram.compute(self.data)

        branches = [s for s in d.all_structures if s.is_branch]
        leaves = [s for s in d.all_structures if s.is_leaf]

        assert len(branches) == 2
        assert len(leaves) == 3

    def test_position_criterion(self):

        def position(structure, index=None, value=None):
            return (np.any(structure.indices()[0] == 6) or
                    np.any(structure.indices()[0] == 8))

        d = Dendrogram.compute(self.data, is_independent=position)

        branches = [s for s in d.all_structures if s.is_branch]
        leaves = [s for s in d.all_structures if s.is_leaf]

        assert len(branches) == 1
        assert len(leaves) == 2

        # Check that leaf that used to contain pixels 1, 2, and 3 is now just
        # part of the main branch.
        assert np.all(branches[0].indices(subtree=False) == np.array([0, 1, 2, 3, 4, 7, 9]))


# Try and reproduce the benchmark tests using a function instead of arguments


from .build_benchmark import BENCHMARKS


@pytest.mark.parametrize(('filename'), BENCHMARKS.keys())
def test_benchmark(filename):

    from astropy.io import fits
    import os

    path = os.path.join(os.path.dirname(__file__),
                        'benchmark_data', filename)

    p = BENCHMARKS[filename]
    data = fits.getdata(path, 1)

    # Now define a function that will test the criteria
    def is_independent_test(structure, index=None, value=None):
        if value is None:
            value = structure.vmin
        if 'min_delta' in p:
            if structure.vmax - value < p['min_delta']:
                return False
        if 'min_npix' in p:
            if len(structure.values()) < p['min_npix']:
                return False
        return True

    d1 = Dendrogram.compute(data,
                            is_independent=is_independent_test,
                            min_value=p['min_value'] if 'min_value' in p else -np.inf)
    d2 = Dendrogram.load_from(path)

    assert d1 == d2



# Licensed under an MIT open source license - see LICENSE
"""
The pruning module provides several functions to perform common
pruning via the ``is_independent`` keyword in the Dendrogram
:meth:`~astrodendro.dendrogram.Dendrogram.compute` method.

Examples::

    #prune unless leaf peak value >= 5
    Dendrogram.compute(data, is_independent=min_peak(5))

    #prune unless leaf contains 10 pixels
    Dendrogram.compute(data, is_independent=min_npix(10))

    #apply both criteria
    is_independent = all_true((min_peak(5), min_npix(10)))
    Dendrogram.compute(data, is_independent=is_independent)
"""

import numpy as np


def _ravel_multi_index(multi_index, dims, mode='raise'):
    #partial implementation of ravel_multi_index,
    #for compatibility with numpy <= 1.5
    #does not implement order kwarg
    ndim = len(dims)

    if len(multi_index) != len(dims):
        raise ValueError("parameter multi_index must be "
                         "a sequence of length %i" % ndim)

    indices = [np.asarray(m) for m in multi_index]
    if mode == 'raise':
        for i, d in zip(indices, dims):
            if ((i < 0) | (i >= d)).any():
                raise ValueError("invalid entry in coordinates array")
    elif mode == 'clip':
        indices = [np.clip(i, 0, d - 1) for i, d in zip(indices, dims)]
    else:  # mode == 'wrap'
        indices = [i % d for i, d in zip(indices, dims)]

    result = np.zeros(len(multi_index[0]), dtype=np.int)
    offset = 1
    for i, d in list(zip(indices, dims))[::-1]:
        result += (i * offset).ravel()
        offset *= d

    return result


if not hasattr(np, 'ravel_multi_index'):
    np.ravel_multi_index = _ravel_multi_index


def all_true(funcs):
    """Combine several ``is_independent`` functions into one

    Parameters
    ----------
    funcs : list-like
        A list of ``is_independent`` functions

    Returns
    -------
    combined_func : function
        A new function which returns true of all the input functions are true
    """
    def result(*args, **kwargs):
        return all(f(*args, **kwargs) for f in funcs)
    return result


def min_delta(delta):
    """
    Minimum delta criteria

    Parameters
    ----------
    delta : float
        The minimum height of a leaf above its merger level

    """
    def result(structure, index=None, value=None):
        if value is None:
            return (structure.vmax - structure.vmin) >= delta
        return (structure.vmax - value) >= delta
    return result


def min_sum(sum):
    """
    Minimum sum criteria

    Parameters
    ----------
    sum : float
        The minimum sum of the pixel values in a leaf
    """
    def result(structure, index=None, value=None):
        return np.nansum(structure.values()) >= sum
    return result


def min_peak(peak):
    """
    Minimum peak criteria

    Parameters
    ----------
    peak : float
        The minimum peak pixel value in a leaf
    """
    def result(structure, index=None, value=None):
        return structure.vmax >= peak
    return result


def min_npix(npix):
    """
    Minimum npix criteria

    Parameters
    ----------
    npix : int
        The minimum number of pixels in a leaf
    """
    def result(structure, index=None, value=None):
        return len(structure.values()) >= npix
    return result


def contains_seeds(seeds):
    """
    Critieria that leaves contain at least one of a list of seed positions

    Parameters
    ----------
    seeds : tuple of array-like
        seed locations. The ith array in the tuple lists the ith coordinate
        for each seed. This is the format returned, e.g., by np.where
    """
    shp = [np.asarray(s).max() + 2 for s in seeds]
    rav = np.ravel_multi_index(seeds, shp)

    def result(structure, index=None, value=None):
        sid = structure.indices()
        if len(sid) != len(seeds):
            raise TypeError("Dimensions of seeds and data do not agree")
        rav2 = np.ravel_multi_index(sid, shp, mode='clip')
        return np.intersect1d(rav, rav2).size > 0

    return result



import numpy as np

from ..pruning import *
from ..structure import Structure

data = np.array([[0, 0, 0, 0],
                 [0, 2, 1, 0],
                 [0, 1, 0, 0],
                 [0, 0, 0, 0]])
st = Structure(zip(*np.where(data)), data[np.where(data)])

def test_npix():
    assert min_npix(3)(st)
    assert not min_npix(4)(st)

def test_min_delta():
    assert min_delta(1)(st)
    assert not min_delta(1.1)(st)
    assert min_delta(1.1)(st, value=0)


def test_min_peak():
    assert min_peak(2)(st)
    assert not min_peak(2.1)(st)

def test_min_sum():
    assert min_sum(4)(st)
    assert not min_sum(4.1)(st)

def test_contains_seeds():
    assert contains_seeds(np.where(data == 2))(st)
    assert contains_seeds(np.where(data == 1))(st)
    assert contains_seeds(([1], [1]))(st)
    assert not contains_seeds(np.where(data == 0))(st)

def test_all_true():
    c1 = min_npix(3)
    c2 = min_peak(2)
    c3 = min_npix(4)
    assert all_true((c1, c2))(st)
    assert not all_true((c1, c2, c3))(st)

def test_multi_ravel():
    from ..pruning import _ravel_multi_index

    x = _ravel_multi_index([[0, 1], [0, 1]], [3, 3])
    np.testing.assert_array_equal(x, [0, 4])

    x = _ravel_multi_index([[1, 0], [0, 1]], [4, 2])
    np.testing.assert_array_equal(x, [2, 1])

    x = _ravel_multi_index([[0], [9]], [5, 3], mode='clip')
    np.testing.assert_array_equal(x, [2])

    x = _ravel_multi_index([[0], [9]], [5, 3], mode='wrap')
    np.testing.assert_array_equal(x, [0])



# Licensed under an MIT open source license - see LICENSE

# The tests here ensure that the Structure class behaves as expected and has
# the correct interface. This is done by setting up a few simple examples and
# ensuring that all the properties and methods are behaving as expected.

import pytest
import numpy as np
from numpy.testing import assert_allclose
from ..structure import Structure
from .test_index import assert_identical_fancyindex


@pytest.mark.parametrize('index', [(0,), (1, 3), (4, 5, 9)])
def test_init_leaf_scalar(index):

    s = Structure(index, 1.5)

    # Properties
    assert s.idx is None
    assert s.is_leaf
    assert not s.is_branch
    assert_identical_fancyindex(s.indices(subtree=False),
                               tuple(np.atleast_1d(i) for i in index))
    assert np.all(s.indices() == s.indices(subtree=True))
    assert np.all(s.values(subtree=False) == np.array([1.5]))
    assert np.all(s.values(subtree=True) == s.values(subtree=False))
    assert s.vmin == 1.5
    assert s.vmax == 1.5
    assert s.height == 1.5
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == []
    assert s.descendants == []

    # Methods
    for subtree in [False, True]:
        assert s.get_npix(subtree=subtree) == 1
        assert s.get_peak(subtree=subtree) == (index, 1.5)

    # Footprint
    array = np.zeros([20 for i in range(len(index))])
    s._fill_footprint(array, level=2)
    assert array[index] == 2
    assert np.sum(array) == 2.


@pytest.mark.parametrize('index', [[(0,), (1,), (2,)],
                                   [(1, 3), (2, 2), (4, 1)],
                                   [(4, 5, 9), (3, 2, 1), (6, 7, 8)]])
def test_init_leaf_list(index):

    s = Structure(index, [3.1, 4.2, 5.3])

    # Properties
    assert s.idx is None
    assert s.is_leaf
    assert not s.is_branch
    indices = tuple(np.atleast_1d(i) for i in zip(*index))
    assert_identical_fancyindex(s.indices(subtree=False), indices)
    assert_identical_fancyindex(s.indices(subtree=True), indices)
    assert np.all(s.values(subtree=False) == np.array([3.1, 4.2, 5.3]))
    assert np.all(s.values(subtree=True) == s.values(subtree=False))
    assert s.vmin == 3.1
    assert s.vmax == 5.3
    assert s.height == 5.3
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == []
    assert s.descendants == []

    # Methods
    for subtree in [False, True]:
        assert s.get_npix(subtree=subtree) == 3
        assert s.get_peak(subtree=subtree) == (index[2], 5.3)

    # Footprint
    array = np.zeros([20 for i in range(len(index[0]))])
    s._fill_footprint(array, level=2)
    for i in index:
        print(i, array[i])
        assert array[i] == 2
    assert np.sum(array) == 6.


@pytest.mark.parametrize('index', [(0,), (1, 3), (4, 5, 9)])
def test_init_branch_scalar(index):

    leaf_index = tuple([10 for i in range(len(index))])
    leaf = Structure(leaf_index, 20.)
    leaf_indices = leaf.indices(subtree=False)

    s = Structure(index, 1.5, children=[leaf])

    # Properties
    assert s.idx is None
    assert not s.is_leaf
    assert s.is_branch
    indices = tuple(np.atleast_1d(i) for i in index)
    indices_all = tuple(np.hstack(a)
                        for a in zip(indices, leaf_indices))
    assert_identical_fancyindex(s.indices(subtree=False), indices)
    assert_identical_fancyindex(s.indices(subtree=True), indices_all)
    assert np.all(s.values(subtree=False) == np.array([1.5]))
    assert np.all(s.values(subtree=True) == np.hstack(([1.5], leaf.values(subtree=False))))
    assert s.vmin == 1.5
    assert s.vmax == 1.5
    assert s.height == 20.
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == [leaf]
    assert s.descendants == [leaf]

    # Leaf properties
    assert leaf.level == 1
    assert leaf.ancestor is s
    assert leaf.parent is s
    assert leaf.children == []
    assert leaf.descendants == []

    # Methods
    assert s.get_npix(subtree=False) == 1
    assert s.get_peak(subtree=False) == (index, 1.5)
    assert s.get_npix(subtree=True) == 2
    assert s.get_peak(subtree=True) == (leaf_index, 20.)

    # Footprint
    array = np.zeros([20 for i in range(len(index))])
    s._fill_footprint(array, level=2)
    assert array[index] == 2.
    assert array[leaf_index] == 3.
    assert np.sum(array) == 5.


@pytest.mark.parametrize('index', [[(0,), (1,), (2,)],
                                   [(1, 3), (2, 2), (4, 1)],
                                   [(4, 5, 9), (3, 2, 1), (6, 7, 8)]])
def test_init_branch_list(index):

    ndim = len(index[0])

    leaf_index = tuple([10 for i in range(ndim)])

    leaf = Structure(leaf_index, 20.)
    leaf_indices = leaf.indices(subtree=False)

    s = Structure(index, [3.1, 4.2, 5.3], children=[leaf])

    # Properties
    assert s.idx is None
    assert not s.is_leaf
    assert s.is_branch
    indices = tuple(np.atleast_1d(i) for i in zip(*index))
    indices_all = tuple(np.hstack(a) for a in
                        zip(indices, leaf_indices))

    assert_identical_fancyindex(s.indices(subtree=False), indices)
    assert_identical_fancyindex(s.indices(subtree=True), indices_all)

    assert np.all(s.values(subtree=False) == np.array([3.1, 4.2, 5.3]))
    assert np.all(s.values(subtree=True) == np.hstack(([3.1, 4.2, 5.3], leaf.values(subtree=False))))
    assert s.vmin == 3.1
    assert s.vmax == 5.3
    assert s.height == 20.
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == [leaf]
    assert s.descendants == [leaf]

    # Leaf properties
    assert leaf.level == 1
    assert leaf.ancestor is s
    assert leaf.parent is s
    assert leaf.children == []
    assert leaf.descendants == []

    # Methods
    assert s.get_npix(subtree=False) == 3
    assert s.get_peak(subtree=False) == (index[2], 5.3)
    assert s.get_npix(subtree=True) == 4
    assert s.get_peak(subtree=True) == (leaf_index, 20.)

    # Footprint
    array = np.zeros([20 for i in range(ndim)])
    s._fill_footprint(array, level=2)
    for i in index:
        assert array[i] == 2.
    assert array[leaf_index] == 3.
    assert np.sum(array) == 9.


@pytest.mark.parametrize('index', [(0,), (1, 3), (4, 5, 9)])
def test_init_branch_scalar_3_level(index):

    leaf_index = tuple([10 for i in range(len(index))])
    leaf = Structure(leaf_index, 20.)
    leaf_indices = leaf.indices(subtree=False)

    branch_index = tuple([9 for i in range(len(index))])
    branch = Structure(branch_index, 15., children=[leaf])
    branch_indices = branch.indices(subtree=False)

    s = Structure(index, 1.5, children=[branch])

    # Properties
    assert s.idx is None
    assert not s.is_leaf
    assert s.is_branch

    indices = tuple(np.atleast_1d(i) for i in index)
    indices_all = tuple(np.hstack(a)
                        for a in zip(indices, branch_indices, leaf_indices))
    assert_identical_fancyindex(s.indices(subtree=False), indices)
    assert_identical_fancyindex(s.indices(subtree=True), indices_all)

    assert np.all(s.values(subtree=False) == np.array([1.5]))
    assert np.all(s.values(subtree=True) == np.hstack((s.values(subtree=False), branch.values(subtree=False), leaf.values(subtree=False))))
    assert s.vmin == 1.5
    assert s.vmax == 1.5
    assert s.height == 15.
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == [branch]
    assert set(s.descendants) == set([branch, leaf])  # order should not matter

    # Branch properties
    assert branch.level == 1
    assert branch.ancestor is s
    assert branch.parent is s
    assert branch.children == [leaf]
    assert branch.descendants == [leaf]

    # Leaf properties
    assert leaf.level == 2
    assert leaf.ancestor is s
    assert leaf.parent is branch
    assert leaf.children == []
    assert leaf.descendants == []

    # Methods
    assert s.get_npix(subtree=False) == 1
    assert s.get_peak(subtree=False) == (index, 1.5)
    assert s.get_npix(subtree=True) == 3
    assert s.get_peak(subtree=True) == (leaf_index, 20.)

    # Footprint
    array = np.zeros([20 for i in range(len(index))])
    s._fill_footprint(array, level=2)
    assert array[index] == 2.
    assert array[branch_index] == 3.
    assert array[leaf_index] == 4.
    assert np.sum(array) == 9.


@pytest.mark.parametrize('index', [[(0,), (1,), (2,)],
                                   [(1, 3), (2, 2), (4, 1)],
                                   [(4, 5, 9), (3, 2, 1), (6, 7, 8)]])
def test_init_branch_list_3_level(index):

    ndim = len(index[0])

    leaf_index = tuple([10 for i in range(ndim)])
    leaf = Structure(leaf_index, 20.)
    leaf_indices = leaf.indices(subtree=False)

    branch_index = tuple([9 for i in range(ndim)])
    branch = Structure(branch_index, 15., children=[leaf])
    branch_indices = branch.indices(subtree=False)

    s = Structure(index, [3.1, 4.2, 5.3], children=[branch])

    # Properties
    assert s.idx is None
    assert not s.is_leaf
    assert s.is_branch

    indices = tuple(np.atleast_1d(i) for i in zip(*index))
    indices_all = tuple(np.hstack(a) for a in
                        zip(indices, branch_indices, leaf_indices))
    assert_identical_fancyindex(s.indices(subtree=False), indices)
    assert_identical_fancyindex(s.indices(subtree=True), indices_all)

    assert np.all(s.values(subtree=False) == np.array([3.1, 4.2, 5.3]))
    assert np.all(s.values(subtree=True) == np.hstack((s.values(subtree=False), branch.values(subtree=False), leaf.values(subtree=False))))
    assert s.vmin == 3.1
    assert s.vmax == 5.3
    assert s.height == 15.
    assert s.level == 0
    assert s.ancestor is s
    assert s.parent is None
    assert s.children == [branch]
    assert s.descendants == [branch, leaf]

    # Branch properties
    assert branch.level == 1
    assert branch.ancestor is s
    assert branch.parent is s
    assert branch.children == [leaf]
    assert branch.descendants == [leaf]

    # Leaf properties
    assert leaf.level == 2
    assert leaf.ancestor is s
    assert leaf.parent is branch
    assert leaf.children == []
    assert leaf.descendants == []

    # Methods
    assert s.get_npix(subtree=False) == 3
    assert s.get_peak(subtree=False) == (index[2], 5.3)
    assert s.get_npix(subtree=True) == 5
    assert s.get_peak(subtree=True) == (leaf_index, 20.)

    # Footprint
    array = np.zeros([20 for i in range(ndim)])
    s._fill_footprint(array, level=2)
    for i in index:
        assert array[i] == 2.
    assert array[branch_index] == 3.
    assert array[leaf_index] == 4.
    assert np.sum(array) == 13.


def test_add_pixel():

    s = Structure(1, 10.)

    assert s.get_npix() == 1
    assert s.get_peak() == (1, 10)
    assert s.vmin == 10.
    assert s.vmax == 10.

    s._add_pixel(2, 8.)

    assert s.get_npix() == 2
    assert s.get_peak() == (1, 10)
    assert s.vmin == 8.
    assert s.vmax == 10.

    s._add_pixel(3, 12.)

    assert s.get_npix() == 3
    assert s.get_peak() == (3, 12.)
    assert s.vmin == 8.
    assert s.vmax == 12.

def test_sorted_leaves():
    l1 = Structure(1, 10., idx=1)
    l2 = Structure(2, 8., idx=2)
    s = Structure(3, 5., children=[l1, l2], idx=3)
    assert s.sorted_leaves() == [l2, l1]
    assert s.sorted_leaves(reverse=True) == [l1, l2]
    def key(x):
        return x.idx
    assert s.sorted_leaves(sort_key=key) == [l1, l2]

    s2 = Structure(4, 3., children=[s], idx=4)
    assert s2.sorted_leaves() == [l2, l1]
    assert s2.sorted_leaves(subtree=False) == []

# TODO: add newick tests



# Licensed under an MIT open source license - see LICENSE

# Notes:
# - A structure is a Leaf or a Branch
# - An ancestor is the largest structure that a structure is part of

import numpy as np

from .structure import Structure
from .progressbar import AnimatedProgressBar
from .io import IO_FORMATS
from . import pruning
from . import six


def _sorted_by_idx(d):
    return sorted(d, key=lambda s: s.idx)


class Dendrogram(object):
    """
    This class is used to compute and represent a dendrogram for a given dataset.

    To create a dendrogram from an array, use the
    :meth:`~astrodendro.dendrogram.Dendrogram.compute` class method::

        >>> from astrodendro import Dendrogram
        >>> d = Dendrogram.compute(array)

    Once the dendrogram has been computed, you can explore it programmatically
    using the ``trunk`` attribute, which allows you to access the base-level
    structures in the dendrogram::

        >>> d.trunk
        [<Structure type=leaf idx=101>,
         <Structure type=branch idx=2152>,
         <Structure type=leaf idx=733>,
         <Structure type=branch idx=303>]

    Structures can then be recursively explored. For more information on
    attributes and methods available for structures, see the
    :class:`~astrodendro.structure.Structure` class.

    The dendrogram can also be explored using an interactive viewer. To use
    this, use the :meth:`~astrodendro.dendrogram.Dendrogram.viewer` method::

        >>> d.viewer()

    and an interactive Matplotlib window should open.

    Finally, the :meth:`~astrodendro.dendrogram.Dendrogram.plotter` method can
    be used to facilitate the creation of plots:

        >>> p = d.plotter()

    For more information on using the plotter and other aspects of the
    :class:`~astrodendro.dendrogram.Dendrogram` class, see the online
    documentation.
    """

    def __init__(self):
        self.data = None
        self.n_dim = 0
        # Put in a friendly error message to make sure nobody confuses the
        # static methods for creating a dendrogram with instance methods:

        def static_warning(self, *args, **kwargs):
            err = "Invalid use of static method. Try d=Dendrogram.compute(data)"
            err += " or d=Dendrogram.load_from(file)"
            raise AttributeError(err)
        self.compute = static_warning
        self.load_from = static_warning

    @staticmethod
    def compute(data, min_value=-np.inf, min_delta=0, min_npix=0,
                is_independent=None, verbose=False):
        """
        Compute a dendrogram from a Numpy array.

        Parameters
        ----------
        data : `~numpy.ndarray`
            The n-dimensional array to compute the dendrogram for
        min_value : float, optional
            The minimum data value to go down to when computing the
            dendrogram. Values below this threshold will be ignored.
        min_delta : float, optional
            The minimum height a leaf has to have in order to be considered an
            independent entity.
        min_npix : int, optional
            The minimum number of pixels/values needed for a leaf to be considered
            an independent entity.
        is_independent : function or list of functions, optional
            A custom function that can be specified that will determine if a
            leaf can be treated as an independent entity. The signature of the
            function should be ``func(structure, index=None, value=None)``
            where ``structure`` is the structure under consideration, and
            ``index`` and ``value`` are optionally the pixel that is causing
            the structure to be considered for merging into/attaching to the
            tree.

            If multiple functions are provided as a list, they
            are all applied when testing for independence.

        Examples
        --------

        The following example demonstrates how to compute a dendrogram from an
        dataset contained in a FITS file::

            >>> from astropy.io import fits
            >>> array = fits.getdata('observations.fits')
            >>> from astrodendro import Dendrogram
            >>> d = Dendrogram.compute(array)

        Notes
        -----
        More information about the above parameters is available from the
        online documentation at [www.dendrograms.org](www.dendrograms.org).
        """
        tests = [pruning.min_delta(min_delta),
                 pruning.min_npix(min_npix)]
        if is_independent is not None:
            if hasattr(is_independent, '__iter__'):
                tests.extend(is_independent)
            else:
                tests.append(is_independent)
        is_independent = pruning.all_true(tests)

        self = Dendrogram()
        self.data = data
        self.n_dim = len(data.shape)
        # For reference, store the parameters used:
        self.params = dict(min_npix=min_npix, min_value=min_value,
                           min_delta=min_delta)

        # Create a list of all points in the cube above min_value
        keep = self.data > min_value
        data_values = self.data[keep]
        indices = np.vstack(np.where(keep)).transpose()

        if verbose:
            print("Generating dendrogram using {:,} of {:,} pixels ({}% of data)".format(data_values.size, self.data.size, (100 * data_values.size / self.data.size)))
            progress_bar = AnimatedProgressBar(end=max(data_values.size, 1), width=40, fill='=', blank=' ')

        # Define index array indicating what structure each cell is part of
        # We expand each dimension by one, so the last value of each
        # index (accessed with e.g. [nx,#,#] or [-1,#,#]) is always zero
        # This permits an optimization below when finding adjacent structures
        self.index_map = -np.ones(np.add(self.data.shape, 1), dtype=np.int32)

        # Dictionary of currently-defined structures:
        structures = {}

        # Define a list of offsets we add to any coordinate to get the indices
        # of all neighbouring pixels
        if self.n_dim == 3:
            neighbour_offsets = np.array([(0, 0, -1), (0, 0, 1), (0, -1, 0), (0, 1, 0), (-1, 0, 0), (1, 0, 0)])
        elif self.n_dim == 2:
            neighbour_offsets = np.array([(0, -1), (0, 1), (-1, 0), (1, 0)])
        elif self.n_dim == 1:
            neighbour_offsets = np.array([(-1, ), (1, ), ])
        else:  # N-dimensional case. Analogous to the above.
            neighbour_offsets = np.concatenate((
                np.identity(self.n_dim, dtype=int),
                np.identity(self.n_dim, dtype=int) * -1
            ))

        # Loop from largest to smallest data_value value. Each time, check if
        # the pixel connects to any existing leaf. Otherwise, create new leaf.

        count = 0

        for i in np.argsort(data_values)[::-1]:

            def next_idx():
                return i + 1
                # Generate IDs index i. We add one to avoid ID 0

            data_value = data_values[i]
            coord = tuple(indices[i])

            # Print stats
            count += 1
            if verbose and (count % 100 == 0):
                progress_bar + 100
                progress_bar.show_progress()

            # Check if point is adjacent to any leaf
            # We don't worry about the edges, because overflow or underflow in
            # any one dimension will always land on an extra "padding" cell
            # with value zero added above when index_map was created
            indices_adjacent = [tuple(c) for c in np.add(neighbour_offsets, indices[i])]
            adjacent = [self.index_map[c] for c in indices_adjacent if self.index_map[c] > -1]

            # Replace adjacent elements by its ancestor
            adjacent = [structures[a].ancestor for a in adjacent]

            # Remove duplicates
            adjacent = _sorted_by_idx(set(adjacent))

            # What happens next depends on how many unique adjacent structures there are

            if not adjacent:  # No adjacent structures;  Create new leaf:

                # Set absolute index of the new element
                idx = next_idx()

                # Create leaf
                leaf = Structure(coord, data_value, idx=idx, dendrogram=self)

                # Add leaf to overall list
                structures[idx] = leaf

                # Set absolute index of pixel in index map
                self.index_map[coord] = idx

            elif len(adjacent) == 1:  # Add to existing leaf or branch

                # Add point to structure
                adjacent[0]._add_pixel(coord, data_value)

                # Set absolute index of pixel in index map
                self.index_map[coord] = adjacent[0].idx

            else:  # Merge leaves

                # At this stage, the adjacent structures might consist of an
                # arbitrary number of leaves and branches.

                # Find all leaves that are not important enough to be
                # kept separate. These leaves will now be treated the
                # same as the pixel under consideration
                merge = [structure for structure in adjacent
                         if structure.is_leaf and
                         (structure.vmax == data_value or
                          not is_independent(structure, index=coord,
                                            value=data_value))]

                # Remove merges from list of adjacent structures
                for structure in merge:
                    adjacent.remove(structure)

                # Now, figure out what object this pixel belongs to
                # How many significant adjacent structures are left?

                if not adjacent:  # if len(adjacent) == 0:
                    # There are no separate leaves left (and no branches), so pick the
                    # first one as the reference and merge all the others onto it
                    belongs_to = merge.pop()
                    belongs_to._add_pixel(coord, data_value)
                elif len(adjacent) == 1:
                    # There is one significant adjacent leaf/branch left.
                    belongs_to = adjacent[0]
                    belongs_to._add_pixel(coord, data_value)
                else:
                    # Create a branch
                    belongs_to = Structure(coord, data_value,
                                           children=adjacent, idx=next_idx(),
                                           dendrogram=self)
                    # Add branch to overall list
                    structures[belongs_to.idx] = belongs_to

                # Set absolute index of pixel in index map
                self.index_map[coord] = belongs_to.idx

                # Add all insignificant leaves in 'merge' to the same object as this pixel:
                for m in merge:
                    # print "Merging leaf %i onto leaf %i" % (i, idx)
                    # Remove leaf
                    structures.pop(m.idx)
                    # Merge the insignificant structure that this pixel now belongs to:
                    belongs_to._merge(m)
                    # Update index map
                    m._fill_footprint(self.index_map, belongs_to.idx)

        if verbose:
            progress_bar.progress = 100  # Done
            progress_bar.show_progress()
            print("")  # newline

        # Create trunk from objects with no ancestors
        self.trunk = _sorted_by_idx([structure for structure in six.itervalues(structures) if structure.parent is None])

        # Remove orphan leaves that aren't large enough
        leaves_in_trunk = [structure for structure in self.trunk if structure.is_leaf]
        for leaf in leaves_in_trunk:
            if not is_independent(leaf):
                # This leaf is an orphan, so remove all references to it:
                structures.pop(leaf.idx)
                self.trunk.remove(leaf)
                leaf._fill_footprint(self.index_map, -1)

        # To make the structure.level property fast, we ensure all the structures in the
        # trunk have their level cached as "0"
        for structure in self.trunk:
            structure._level = 0  # See the definition of level() in structure.py

        # Save a list of all structures accessible by ID
        self._structures_dict = {}

        # Re-assign idx and update index map
        sorted_structures = sorted(self, key=lambda s: s.smallest_index)
        for idx, s in enumerate(sorted_structures):
            s.idx = idx
            s._fill_footprint(self.index_map, idx, recursive=False)
            self._structures_dict[idx] = s

        # Remove border from index map
        s = tuple(slice(0, s, 1) for s in data.shape)
        self.index_map = self.index_map[s]

        # Add dendrogram index
        self._index()

        # Return the newly-created dendrogram:
        return self


    def _index(self):
        # add dendrogram index
        ti = TreeIndex(self)

        for s in six.itervalues(self._structures_dict):
            s._tree_index = ti


    @property
    def trunk(self):
        """
        A list of all structures that have no parent structure and form the
        base of the tree.
        """
        return self._trunk

    @trunk.setter
    def trunk(self, value):
        self._trunk = value

    @staticmethod
    def load_from(filename, format=None):
        """
        Load a previously computed dendrogram from a file.

        Parameters
        ----------
        filename : str
            The name of the file to load the dendrogram from. By default, the
            file format will be automatically detected from the file
            extension. At this time, only HDF5 files (extension ``.hdf5``) are
            supported.
        format : str, optional
            The format to use to read the file. By default, this is not used
            and the format is auto-detected from the file extension. At this
            time, the only format supported is ``'hdf5'``.
        """
        from .io import load_dendrogram
        return load_dendrogram(filename, format=format)

    def save_to(self, filename, format=None):
        """
        Save the dendrogram to a file.

        Parameters
        ----------
        filename : str
            The name of the file to save the dendrogram to. By default, the
            file format will be automatically detected from the file
            extension. At this time, only HDF5 files (extension ``.hdf5``) are
            supported.
        format : str, optional
            The format to use for the file. By default, this is not used and
            the format is auto-detected from the file extension. At this time,
            the only format supported is ``'hdf5'``.
        """
        from .io import save_dendrogram
        return save_dendrogram(self, filename, format=format)

    @property
    def leaves(self):
        """
        A flattened list of all leaves in the dendrogram.
        """
        return [i for i in six.itervalues(self._structures_dict) if i.is_leaf]

    def to_newick(self):
        #this caches newicks, and prevents too much recursion
        [s.newick for s in reversed(list(self.all_structures))]

        return "(%s);" % ','.join([structure.newick for structure
                                   in self.trunk])

    def structure_at(self, indices):
        """
        Get the structure at the specified pixel coordinate.

        This will return None if no structure includes the specified pixel
        coordinates.
        """
        idx = self.index_map[indices]
        if idx > -1:
            return self._structures_dict[idx]
        return None

    @property
    def all_structures(self):
        """
        Yields an iterator over all structures in the dendrogram, in prefix order.
        """

        todo = list(self.trunk)
        while len(todo) > 0:
            st = todo.pop(0)
            yield st
            todo = st.children + todo

    def __getitem__(self, key):
        """Fetch structures by index value"""
        return self._structures_dict[key]

    def __len__(self):
        """Return number of structures in dendrogram"""
        return len(self._structures_dict)

    def __iter__(self):
        return self.all_structures

    def __eq__(self, other):
        if not isinstance(other, Dendrogram):
            return False

        if not (self.data == other.data).all():
            return False

        # structures should have the same extent,
        # but idx values need not be identical. This
        # tests the index map for that
        u, ind = np.unique(self.index_map, return_index=True)
        u, ind2 = np.unique(self.index_map, return_index=True)
        return (np.sort(ind) == np.sort(ind2)).all()

    def plotter(self):
        """
        Return a :class:`~astrodendro.plot.DendrogramPlotter` instance that makes it easier to construct plots.
        """
        from .plot import DendrogramPlotter
        return DendrogramPlotter(self)

    def viewer(self):
        """
        Launch an interactive viewer to explore the dendrogram.

        This functionality is only available for 2- or 3-d datasets.
        """
        from .viewer import BasicDendrogramViewer
        return BasicDendrogramViewer(self)


class TreeIndex(object):
    def __init__(self, dendrogram):
        """
        Object that efficiently extracts the locations of Structures in an
        ndarray.

        Parameters
        ----------
        dendrogram : Dendrogram instance
                    The dendrogram to index
        """
        index_map = dendrogram.index_map
        trunk = dendrogram.trunk

        sz = index_map.size
        nd = len(index_map.shape)

        assert sz == dendrogram.data.size
        assert index_map.min() >= -1

        #map ids to [0, 1, ...] for storage efficiency
        uniq, bins = np.unique(index_map, return_inverse=True)
        packed = dict((u, i) for i, u in enumerate(uniq))

        flat_idx = index_map.ravel()
        ri = np.argsort(bins)
        idx_ct = np.bincount(bins)
        idx_sub_ct = {}
        idx_cdf = np.hstack((0, np.cumsum(idx_ct)))

        #efficiently build up npix values
        structures = reversed(sorted(dendrogram._structures_dict.values(),
                                key=lambda x: x.level))
        for st in structures:
            idx_sub_ct[st.idx] = idx_ct[packed[st.idx]]
            idx_sub_ct[st.idx] += sum(idx_sub_ct[c.idx] for c in st.children)

        #build a 1D index array with the following properties
        # - values in index reference locations in flattened index_map
        # - every structure (+ subtree) is a continuous slice of index
        # - index[offset[i]] is the first location for (packed) structure i
        # - npix[i] gives number of pixels in (packed) structure i,
        #   exluding subtree
        # - npix_subtree[i] is like above, but includes subtrees
        #
        # In summary, the locations in the flattened_index map
        # for structure i excluding subrees is
        #    pi = packed[i]
        #    index[offset[pi] : offset[pi] + npix[pi]]
        # and including subtrees is
        #    index[offset[pi] : offset[pi] + npix_subtree[pi]]
        offset = np.zeros(idx_ct.size, dtype=np.int)
        npix = offset * 0
        npix_subtree = offset * 0

        index = -np.ones(sz, dtype=np.int)
        order = dendrogram.all_structures

        pos = 0
        for o in order:
            sid = packed[o.idx]
            offset[sid] = pos
            npix[sid] = idx_ct[sid]
            npix_subtree[sid] = idx_sub_ct[o.idx]
            idx = ri[idx_cdf[sid]: idx_cdf[sid] + npix[sid]]
            assert (flat_idx[idx] == o.idx).all()
            index[pos: pos + npix[sid]] = idx
            pos += npix[sid]

        #turn inds back into an ndim index
        self._index = tuple(n.ravel()[index] for n in
                            np.indices(index_map.shape))

        self._data = dendrogram.data
        self._offset = offset
        self._npix = npix
        self._npix_subtree = npix_subtree
        self.packed = packed

    def indices(self, sid, subtree=True):
        """
        Return pixel indices associated with a dendrogram structure.

        Returns the pixels in the original dendrogram array
        which are associated with a particular structure id.

        Parameters
        ----------
        sid : integer
              The structure index to lookup. Stored in `Structure.idx`
        subtree : bool, optional
              If true, return indices for subtrees as well. Default=False

        Returns
        -------
        A tuple of integer ndarrays, akin to np.where().
        This can be directly used as an index into the dendrogram
        data array
        """
        sid = self.packed[sid]
        i0 = self._offset[sid]
        di = self._npix_subtree[sid] if subtree else self._npix[sid]
        return tuple(ind[i0: i0 + di] for ind in self._index)

    def values(self, sid, subtree=True):
        return self._data[self.indices(sid, subtree=subtree)]



# Licensed under an MIT open source license - see LICENSE

import pytest
from mock import patch

import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.wcs import WCS


from ._testdata import data
from ..analysis import (ScalarStatistic, PPVStatistic, ppv_catalog,
                        Metadata, PPStatistic, pp_catalog)
from .. import Dendrogram
from ..structure import Structure


def assert_allclose_quantity(a, b):
    if not isinstance(a, u.Quantity):
        raise TypeError("a is not a quantity")
    if not isinstance(b, u.Quantity):
        raise TypeError("b is not a quantity")
    assert_allclose(a.value, b.value)
    assert a.unit == b.unit

wcs_2d = WCS(header=dict(cdelt1=1, crval1=0, crpix1=1,
                         cdelt2=2, crval2=0, crpix2=1))
wcs_3d = WCS(header=dict(cdelt1=1, crval1=0, crpix1=1,
                         cdelt2=2, crval2=0, crpix2=1,
                         cdelt3=3, crval3=0, crpix3=1))

def benchmark_stat():
    x = np.array([216, 216, 216, 216, 216, 217, 216,
                  216, 216, 217, 216, 217, 218, 216,
                  217, 216, 216, 217, 216, 217, 216])
    y = np.array([48, 50, 51, 47, 48, 48, 49, 50,
                  46, 46, 47, 47, 47, 48, 48, 49,
                  46, 46, 47, 46, 47])
    z = np.array([11, 11, 11, 12, 12, 12, 12, 12,
                  13, 13, 13, 13, 13, 13, 13, 13,
                  14, 14, 14, 15, 15])
    indices = (z, y, x)
    values = data[indices]
    return ScalarStatistic(values, indices)


def benchmark_values():
    result = {}
    result['mom0'] = 45.8145142
    result['mom1'] = [12.8093996, 47.6574211, 216.3799896]
    result['mom2'] = [
        [1.2831592097072804, -1.1429260442312184, 0.1653071908000249],
        [-1.1429260442312184, 2.0107706426889038, -0.2976682802749759],
        [0.1653071908000249, -0.2976682802749759, 0.3306051333187136]]
    result['mom2_100'] = 1.2831592097072804
    result['mom2_010'] = 2.0107706426889038
    result['mom2_011'] = 0.8730196376110217

    result['paxis1'] = [0.5833025351190873,
                        -0.8016446861607931, 0.1308585101314019]
    result['paxis2'] = [0.8038500408313988,
                        0.5466096691013617, -0.2346124069614796]
    result['paxis3'] = [-0.1165472624260408,
                        -0.2420406304632855, -0.9632409194100563]
    result['sig_maj'] = np.sqrt(2.0619485)
    result['sig_min'] = np.sqrt(0.27942730)
    result['area_exact_pp'] = 21
    result['area_exact_ppv'] = 10

    return result


class TestScalarStatistic(object):

    def setup_method(self, method):
        self.stat = benchmark_stat()

    def test_mom0(self):
        stat = self.stat
        assert_allclose(stat.mom0(), benchmark_values()['mom0'])

    def test_mom1(self):
        stat = self.stat
        assert_allclose(stat.mom1(), benchmark_values()['mom1'])

    def test_mom2(self):
        stat = self.stat
        assert_allclose(stat.mom2(), benchmark_values()['mom2'])

    def test_mom2_along(self):
        stat = self.stat
        v = benchmark_values()
        assert_allclose(stat.mom2_along([1, 0, 0]), v['mom2_100'])
        assert_allclose(stat.mom2_along([2, 0, 0]), v['mom2_100'])
        assert_allclose(stat.mom2_along([0, 1, 0]), v['mom2_010'])
        assert_allclose(stat.mom2_along([0, 1, 1]), v['mom2_011'])

    def test_count(self):
        stat = self.stat
        assert_allclose(stat.count(), 21)

    def test_sky_paxes(self):
        stat = self.stat
        v1, v2, v3 = stat.paxes()
        v = benchmark_values()

        #doesn't matter if v = vex * -1.
        assert_allclose(np.abs(np.dot(v1, v['paxis1'])), 1)
        assert_allclose(np.abs(np.dot(v2, v['paxis2'])), 1)
        assert_allclose(np.abs(np.dot(v3, v['paxis3'])), 1)

    def test_projected_paxes(self):
        stat = self.stat
        v = benchmark_values()
        v1, v2 = stat.projected_paxes([[0, 1, 0], [0, 0, 1]])

        assert_allclose(stat.mom2_along([0, v1[0], v1[1]]), v['sig_maj'] ** 2)
        assert_allclose(stat.mom2_along([0, v2[0], v2[1]]), v['sig_min'] ** 2)

    def test_projected_paxes_int(self):
        ind = (np.array([0, 1, 2]),
               np.array([0, 1, 2]),
               np.array([0, 1, 2]))
        v = np.array([1, 1, 1])
        stat = ScalarStatistic(v, ind)
        a, b = stat.projected_paxes([[0, 1, 0], [0, 0, 1]])
        assert_allclose(a, [1 / np.sqrt(2), 1 / np.sqrt(2)])


class TestScalar2D(object):

    def setup_method(self, method):
        x = np.array([213, 213, 214, 211, 212, 212])
        y = np.array([71, 71, 71, 71, 71, 71])
        z = np.array([46, 45, 45, 44, 46, 43])

        ind = (z, y, x)
        val = data[ind].copy()
        ind = (z, x)
        val[0] = 0  # we'll replace this with nan in a subclass

        self.stat = ScalarStatistic(val, ind)

    def test_mom0(self):
        assert_allclose(self.stat.mom0(), 19.2793083)

    def test_mom1(self):
        assert_allclose(self.stat.mom1(), [44.6037369, 212.4050598])

    def test_mom2(self):
        assert_allclose(self.stat.mom2(),
                        [[1.0321983103326871, 0.3604276691031663],
                        [0.3604276691031663, 1.0435691076146387]])

    def test_mom2_along(self):
        assert_allclose(self.stat.mom2_along([0, 1]), 1.0435691076146387)

    def test_count(self):
        assert_allclose(self.stat.count(), 6)

    def test_sky_paxes(self):
        v1, v2 = self.stat.paxes()
        v1ex = [0.7015083489012299, 0.7126612353859795]
        v2ex = [0.7126612353859795, -0.7015083489012299]

        assert_allclose(np.abs(np.dot(v1, v1ex)), 1)
        assert_allclose(np.abs(np.dot(v2, v2ex)), 1)


class TestScalarNan(TestScalar2D):

    def setup_method(self, method):
        x = np.array([213, 213, 214, 211, 212, 212])
        y = np.array([71, 71, 71, 71, 71, 71])
        z = np.array([46, 45, 45, 44, 46, 43])

        ind = (z, y, x)
        val = data[ind].copy()
        ind = (z, x)
        #all tests should be the same as superclass,
        #since nan should = 0 weight
        val[0] = np.nan

        self.stat = ScalarStatistic(val, ind)


class TestPPVStatistic(object):

    def setup_method(self, method):
        self.stat = benchmark_stat()
        self.v = benchmark_values()

    def metadata(self, **kwargs):
        result = dict(data_unit=u.Jy, wcs=wcs_3d)
        result.update(**kwargs)
        return result

    def test_x_cen(self):

        p = PPVStatistic(self.stat, self.metadata())
        assert_allclose(p.x_cen, self.v['mom1'][2])

        p = PPVStatistic(self.stat, self.metadata(vaxis=2))
        assert_allclose(p.x_cen, self.v['mom1'][1] * 2)

    def test_y_cen(self):
        p = PPVStatistic(self.stat, self.metadata())
        assert_allclose(p.y_cen, self.v['mom1'][1] * 2)

        p = PPVStatistic(self.stat, self.metadata(vaxis=2))
        assert_allclose(p.y_cen, self.v['mom1'][0] * 3)

    def test_v_cen(self):
        p = PPVStatistic(self.stat, self.metadata())
        assert_allclose(p.v_cen, self.v['mom1'][0] * 3)

        p = PPVStatistic(self.stat, self.metadata(vaxis=2))
        assert_allclose(p.v_cen, self.v['mom1'][2])

    def test_major_sigma(self):
        p = PPVStatistic(self.stat, self.metadata(spatial_scale=2 * u.arcsec))
        assert_allclose_quantity(p.major_sigma, self.v['sig_maj'] * 2 * u.arcsec)

    def test_minor_sigma(self):
        p = PPVStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.minor_sigma, self.v['sig_min'] * 4 * u.arcsec)

    def test_radius(self):
        p = PPVStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.radius, np.sqrt(self.v['sig_min'] * self.v['sig_maj']) * 4 * u.arcsec)

    def test_area_exact(self):
        p = PPVStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.area_exact,  (4 * u.arcsec)**2 * self.v['area_exact_ppv'])

    def test_area_ellipse(self):
        p = PPVStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.area_ellipse,  (4 * u.arcsec)**2 * self.v['sig_min'] * self.v['sig_maj'] * np.pi * (2.3548 * 0.5)**2)

    def test_v_rms(self):
        p = PPVStatistic(self.stat, self.metadata())
        assert_allclose_quantity(p.v_rms, np.sqrt(self.v['mom2_100']) * u.pixel)

        p = PPVStatistic(self.stat, self.metadata(vaxis=1, velocity_scale=10 * u.km / u.s))
        assert_allclose_quantity(p.v_rms, np.sqrt(self.v['mom2_010']) * 10 * u.km / u.s)

    def test_position_angle(self):
        x = np.array([0, 1, 2])
        y = np.array([1, 1, 1])
        z = np.array([0, 1, 2])
        v = np.array([1, 1, 1])

        ind = (z, y, x)
        stat = ScalarStatistic(v, ind)
        p = PPVStatistic(stat, self.metadata())
        assert_allclose_quantity(p.position_angle, 0 * u.degree)

        ind = (z, x, y)
        stat = ScalarStatistic(v, ind)
        p = PPVStatistic(stat, self.metadata())
        assert_allclose_quantity(p.position_angle, 90 * u.degree)

    def test_units(self):
        m = self.metadata(spatial_scale=1 * u.deg, velocity_scale=1 * u.km / u.s,
                          data_unit=1 * u.K, distance=1 * u.pc)
        p = PPVStatistic(self.stat, m)

        assert p.v_rms.unit == u.km / u.s
        assert p.major_sigma.unit == u.deg
        assert p.minor_sigma.unit == u.deg
        assert p.radius.unit == u.deg


class TestPPStatistic(object):

    def setup_method(self, method):
        self.stat = benchmark_stat()
        #this trick essentially collapses along the 0th axis
        #should preserve sky_maj, sky_min
        self.stat.indices = (self.stat.indices[1], self.stat.indices[2])
        self.v = benchmark_values()

    def metadata(self, **kwargs):
        result = dict()
        result.update(**kwargs)
        return result

    def test_major_sigma(self):
        p = PPStatistic(self.stat, self.metadata(spatial_scale=2 * u.arcsec))

        assert_allclose_quantity(p.major_sigma, self.v['sig_maj'] * 2 * u.arcsec)

    def test_minor_sigma(self):
        p = PPStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.minor_sigma, self.v['sig_min'] * 4 * u.arcsec)

    def test_radius(self):
        p = PPStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.radius, np.sqrt(self.v['sig_min'] * self.v['sig_maj']) * 4 * u.arcsec)

    def test_area_exact(self):
        p = PPStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.area_exact, (4 * u.arcsec)**2 * self.v['area_exact_pp'])

    def test_area_ellipse(self):
        p = PPStatistic(self.stat, self.metadata(spatial_scale=4 * u.arcsec))
        assert_allclose_quantity(p.area_ellipse, (4 * u.arcsec)**2 * self.v['sig_min'] * self.v['sig_maj'] * np.pi * (2.3548 * 0.5) ** 2)

    def test_position_angle(self):
        x = np.array([0, 1, 2])
        y = np.array([1, 1, 1])
        v = np.array([1, 1, 1])

        ind = (y, x)
        stat = ScalarStatistic(v, ind)
        p = PPStatistic(stat, self.metadata())
        assert_allclose_quantity(p.position_angle, 0 * u.degree)

        ind = (x, y)
        stat = ScalarStatistic(v, ind)
        p = PPStatistic(stat, self.metadata())
        assert_allclose_quantity(p.position_angle, 90 * u.degree)


def test_statistic_dimensionality():

    d = Dendrogram.compute(np.ones((10,10)))

    with pytest.raises(ValueError) as exc:
        PPVStatistic(d.trunk[0])
    assert exc.value.args[0] == "PPVStatistic can only be used on 3-d datasets"

    PPStatistic(d.trunk[0])

    d = Dendrogram.compute(np.ones((10,10,10)))

    with pytest.raises(ValueError) as exc:
        PPStatistic(d.trunk[0])
    assert exc.value.args[0] == "PPStatistic can only be used on 2-d datasets"

    PPVStatistic(d.trunk[0])


class TestCataloger(object):
    files = []
    cataloger = None

    def stat(self):
        raise NotImplementedError()

    def metadata(self):
        raise NotImplementedError()

    def make_catalog(self, s=None, md=None, fields=None):
        s = s or [self.stat()]
        structures = [Structure(zip(*x.indices), x.values) for x in s]
        md = md or self.metadata()
        return self.cataloger(structures, md, fields)

    def test_benchmark(self):
        c = self.make_catalog()
        assert c.dtype.names == tuple(sorted(self.fields))
        assert len(c) == 1
        c = self.make_catalog(s=[self.stat()] * 3)
        assert len(c) == 3
        return c

    def test_field_selection(self):
        stat = self.stat()
        md = self.metadata()
        c = self.cataloger([Structure(zip(*stat.indices), stat.values)], md, fields=['x_cen'])
        assert c.dtype.names == ('_idx', 'x_cen',)


class TestPPVCataloger(TestCataloger):
    fields = ['_idx', 'flux',
              'major_sigma', 'minor_sigma', 'radius', 'area_ellipse',
              'area_exact', 'v_rms', 'position_angle', 'x_cen', 'y_cen', 'v_cen']
    cataloger = staticmethod(ppv_catalog)

    def stat(self):
        return benchmark_stat()

    def metadata(self):
        return dict(vaxis=1, data_unit=u.Jy, wcs=wcs_3d)


class TestPPCataloger(TestCataloger):
    fields = ['_idx', 'flux',
              'major_sigma', 'minor_sigma', 'radius', 'area_ellipse',
              'area_exact', 'position_angle', 'x_cen', 'y_cen']
    cataloger = staticmethod(pp_catalog)

    def stat(self):
        bs = benchmark_stat()
        bs.indices = (bs.indices[1], bs.indices[2])
        return bs

    def metadata(self):
        return dict(data_unit=u.Jy, wcs=wcs_2d)


#don't let pytest test abstract class
del TestCataloger


def test_metadata_protocol():
    class Foo(object):
        x = Metadata('x', 'test')
        y = Metadata('y', 'test', default=5)
        z = Metadata('z', 'test', strict=True)

        def __init__(self, md):
            self.metadata = md

    f = Foo(dict(x=10))
    assert f.x == 10
    assert f.y == 5
    with pytest.raises(KeyError):
        f.z



# Licensed under an MIT open source license - see LICENSE

import numpy as np
import pytest

from .. import Dendrogram
from ..structure import Structure


class Test2DimensionalData(object):

    def test_dendrogramWithNan(self):

        n = np.nan
        data = np.array([[n, n, n, n, n, n, n, n],
                         [n, 4, n, n, n, n, n, n],
                         [n, n, n, 1, n, n, 0, 5],
                         [3, n, n, 2, 3, 2, 0, n]])
        d = Dendrogram.compute(data)

        ########################################
        # Check the trunk elements:

        leaves = [structure for structure in d.trunk if structure.is_leaf]
        branches = [structure for structure in d.trunk if structure not in leaves]

        assert len(leaves) == 2, "We expect two leaves among the lowest structures (the trunk)"
        assert len(branches) == 1, "We expect one branch among the lowest structures (the trunk)"

        for leaf in leaves:
            assert len(leaf.values(subtree=False)) == 1, "Leaves in the trunk are only expected to contain one point"
            assert leaf.parent is None
            assert leaf.ancestor == leaf
            assert leaf.get_npix() == 1
            if leaf.values(subtree=False)[0] == 4:
                assert list(zip(*leaf.indices(subtree=False)))[0] == (1, 1)
            elif leaf.values(subtree=False)[0] == 3:
                assert list(zip(*leaf.indices(subtree=False)))[0] == (3, 0)
            else:
                self.fail("Invalid value of flux in one of the leaves")

        ########################################
        # Check properties of the branch:
        branch = branches[0]
        assert branch.parent is None
        assert branch.ancestor == branch
        assert branch.get_npix(subtree=False) == 1  # only pixel is a 0
        assert branch.get_npix(subtree=True) == 7

        assert len(branch.children) == 2
        for leaf in branch.children:
            assert leaf.is_leaf
            assert leaf.ancestor == branch
            assert leaf.parent == branch
            if 5 in leaf.values(subtree=False):
                assert sum(leaf.values(subtree=False)) == 5
            elif 3 in leaf.values(subtree=False):
                assert sum(leaf.values(subtree=False)) == 1 + 2 + 3 + 2
            else:
                self.fail("Invalid child of the branch")

    def test_mergeLevelAndHeight(self):
        n = np.nan
        data = np.array([[n, n, n, n, n, ],
                         [n, 4, 2, 5, n, ],
                         [n, n, n, n, 0, ]])
        d = Dendrogram.compute(data)
        branch, leaf4, leaf5 = d.trunk[0], d.structure_at((1, 1)), d.structure_at((1, 3))
        assert leaf4.height == 4.
        assert leaf5.height == 5.
        assert branch.height == 4.

    def test_dendrogramWithConstBackground(self):
        # Test a highly artificial array containing a lot of equal pixels
        data = np.array([[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 3, 5, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 2, 3, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 3, 4, 3, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 2, 3, 2, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 2, 3, 2, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 3, 4, 3, 1, 1, 2, 2, 1, 1, 1],
                         [1, 1, 1, 1, 2, 3, 2, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                         [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], ])
        d = Dendrogram.compute(data)
        assert len(d) <= 7
        # Some of the '1' valued pixels get included with the leaves and branches,
        # hence number of structures is currently 7 and not 6 as expected.
        # Fixing this is probably more trouble than it's worth.
        leaf_with_twos = d.structure_at((10, 9))
        assert leaf_with_twos.height == 2

        # Check that all structures contain a reference to the dendrogram
        for structure in d:
            assert structure._dendrogram is d


class Test3DimensionalData(object):
    def setup_method(self, method):
        from ._testdata import data
        self.data = data

    def test_dendrogramComputation(self):
        d = Dendrogram.compute(self.data, min_npix=8, min_delta=0.3, min_value=1.4)

        # This data with these parameters should produce 55 leaves
        assert len(d.leaves) == 55

        # Now check every pixel in the data cube (this takes a while).
        st_map = -np.ones(self.data.shape, dtype=np.int)
        for st in d.all_structures:
            st_map[st.indices(subtree=False)] = st.idx

        #check that vmin/vmax/peak are correct
        for st in d.all_structures:
            assert st.vmin == self.data[st.indices(subtree=False)].min()
            assert st.vmax == self.data[st.indices(subtree=False)].max()
            pk_exp = self.data[st.indices(subtree=True)].max()

            ind, pk = st.get_peak(subtree=True)
            assert self.data[ind] == pk
            assert pk_exp == pk

        # The "right" way to do this is loop through indices,
        # and repeatedly call structure_at(). However, this is quite slow
        # structure_at is a thin wrapper around index_map,
        # and we compare index_map to st_map instead
        np.testing.assert_array_equal(st_map, d.index_map)

        # here, we test a few values of structure_at
        for coord in np.indices(self.data.shape).reshape(self.data.ndim, np.prod(self.data.shape)).transpose()[::100]:
            coord = tuple(coord)
            f = self.data[coord]
            structure = d.structure_at(coord)
            if structure is not None:
                assert structure.idx == st_map[coord], "Pixel at {0} is claimed to be part of {1}, but that structure does not contain the coordinate {0}!".format(coord, structure)
            else:
                assert st_map[coord] == -1


class TestNDimensionalData(object):
    def test_4dim(self):
        " Test 4-dimensional data "
        data = np.zeros((5, 5, 5, 5))  # Create a 5x5x5x5 array initialized to zero
        # N-dimensional data is hard to conceptualize so I've kept this simple.
        # Create a local maximum (value 5) at the centre
        data[2, 2, 2, 2] = 5
        # add some points around it with value 3. Note that '1:4:2' is equivalent to saying indices '1' and '3'
        data[2, 1:4:2, 2, 2] = data[2, 2, 1:4:2, 2] = data[2, 2, 2, 1:4:2] = 3
        # Add a trail of points of value 2 connecting one of those 3s to a 4
        data[0:3, 0, 2, 2] = 2  # Sets [0, 0, 2, 2], [1, 0, 2, 2], and [2, 0, 2, 2] all equal to 2 -> will connect to the '3' at [2, 1, 2, 2]
        data[0, 0, 2, 1] = 4

        # Now dendrogram it:
        d = Dendrogram.compute(data, min_value=1)
        # We expect two leaves:
        leaves = d.leaves
        assert len(leaves) == 2
        # We expect one branch:
        branches = [i for i in d.all_structures if i.is_branch]
        assert len(branches) == 1
        assert len(d.trunk) == 1
        assert d.trunk[0] == branches[0]

        # The maxima of each leaf should be at [2,2,2,2] and [0,3,2,1]
        for leaf in leaves:
            assert leaf.get_peak() in (((2, 2, 2, 2), 5.), ((0, 0, 2, 1), 4.))
        assert leaves[0].get_peak() != leaves[1].get_peak()

        # Check out a few more properties of the leaf around the global maximum:
        leaf = d.structure_at((2, 2, 2, 2))
        assert leaf.vmax == 5
        assert leaf.vmin == 2
        assert leaf.get_npix() == 1 + 6 + 2  # Contains 1x '5', 6x '3', and 2x '2'. The other '2' should be in the branch
        # Check that the only pixel in the branch is a '2' at [0,0,2,2]
        assert (list(zip(*branches[0].indices(subtree=False))), branches[0].values(subtree=False)) == ([(0, 0, 2, 2), ], [2., ])


from .build_benchmark import BENCHMARKS

@pytest.mark.parametrize(('filename'), BENCHMARKS.keys())
def test_benchmark(filename):
    from astropy.io import fits
    import os

    path = os.path.join(os.path.dirname(__file__),
                        'benchmark_data', filename)
    p = BENCHMARKS[filename]
    data = fits.getdata(path, 1)

    d1 = Dendrogram.compute(data, **p)
    d2 = Dendrogram.load_from(path)

    assert d1 == d2

    # Check that all structures contain a reference to the dendrogram
    for structure in d1:
        assert structure._dendrogram is d1




import numpy as np

from .. import six

def parse_newick(string):

    items = {}

    # Find maximum level
    current_level = 0
    max_level = 0
    for i, c in enumerate(string):
        if c == '(':
            current_level += 1
        if c == ')':
            current_level -= 1
        max_level = max(max_level, current_level)

    # Loop through levels and construct tree
    for level in range(max_level, 0, -1):

        pairs = []

        current_level = 0
        for i, c in enumerate(string):
            if c == '(':
                current_level += 1
                if current_level == level:
                    start = i
            if c == ')':
                if current_level == level:
                    pairs.append((start, i))
                current_level -= 1

        for pair in pairs[::-1]:

            # Extract start and end of branch definition
            start, end = pair

            # Find the ID of the branch
            colon = string.find(":", end)
            branch_id = string[end + 1:colon]
            if branch_id == '':
                branch_id = 'trunk'
            else:
                branch_id = int(branch_id)

            # Add branch definition to overall definition
            items[branch_id] = eval("{%s}" % string[start + 1:end])

            # Remove branch definition from string
            string = string[:start] + string[end + 1:]

    new_items = {}

    def collect(d):
        for item in d:
            if item in items:
                collect(items[item])
                d[item] = (items[item], d[item])
        return

    collect(items['trunk'])

    return items['trunk']


def parse_dendrogram(newick, data, index_map):
    from ..dendrogram import Dendrogram
    from ..structure import Structure

    d = Dendrogram()
    d.ndim = len(data.shape)

    d._structures_dict = {}
    d.data = data
    d.index_map = index_map

    flux_by_structure = {}
    indices_by_structure = {}

    def _construct_tree(repr):
        structures = []
        for idx in repr:
            idx = int(idx)
            structure_indices = indices_by_structure[idx]
            f = flux_by_structure[idx]
            if type(repr[idx]) == tuple:
                sub_structures_repr = repr[idx][0]  # Parsed representation of sub structures
                sub_structures = _construct_tree(sub_structures_repr)
                for i in sub_structures:
                    d._structures_dict[i.idx] = i
                b = Structure(structure_indices, f, children=sub_structures, idx=idx)
                # Correct merge levels - complicated because of the
                # order in which we are building the tree.
                # What we do is look at the heights of this branch's
                # 1st child as stored in the newick representation, and then
                # work backwards to compute the merge level of this branch
                first_child_repr = six.next(six.itervalues(sub_structures_repr))
                if type(first_child_repr) == tuple:
                    height = first_child_repr[1]
                else:
                    height = first_child_repr
                d._structures_dict[idx] = b
                structures.append(b)
            else:
                l = Structure(structure_indices, f, idx=idx)
                structures.append(l)
                d._structures_dict[idx] = l
        return structures

    # Do a fast iteration through d.data, adding the indices and data values
    # to the two dictionaries declared above:
    indices = np.indices(d.data.shape).reshape(d.data.ndim, np.prod(d.data.shape)).transpose()

    for coord in indices:
        coord = tuple(coord)
        idx = d.index_map[coord]
        if idx > -1:
            try:
                flux_by_structure[idx].append(d.data[coord])
                indices_by_structure[idx].append(coord)
            except KeyError:
                flux_by_structure[idx] = [d.data[coord]]
                indices_by_structure[idx] = [coord]

    d.trunk = _construct_tree(parse_newick(newick))
    # To make the structure.level property fast, we ensure all the items in the
    # trunk have their level cached as "0"
    for structure in d.trunk:
        structure._level = 0  # See the @property level() definition in structure.py

    d._index()
    return d



# Licensed under an MIT open source license - see LICENSE

import numpy as np
from .plot import DendrogramPlotter


class BasicDendrogramViewer(object):

    def __init__(self, dendrogram):

        if dendrogram.data.ndim not in [2, 3]:
            raise ValueError("Only 2- and 3-dimensional arrays are supported at this time")

        self.array = dendrogram.data
        self.dendrogram = dendrogram
        self.plotter = DendrogramPlotter(dendrogram)
        self.plotter.sort(reverse=True)

        # Get the lines as individual elements, and the mapping from line to structure
        self.lines = self.plotter.get_lines()

        # Define the currently selected subtree
        self.selected = None
        self.selected_lines = None
        self.selected_contour = None

        # Initiate plot
        import matplotlib.pyplot as plt
        self.fig = plt.figure(figsize=(14, 8))

        self.ax1 = self.fig.add_axes([0.1, 0.1, 0.4, 0.7])

        from matplotlib.widgets import Slider

        self._clim = (np.min(self.array[~np.isnan(self.array) & ~np.isinf(self.array)]),
                      np.max(self.array[~np.isnan(self.array) & ~np.isinf(self.array)]))

        if self.array.ndim == 2:

            self.slice = None
            self.image = self.ax1.imshow(self.array, origin='lower', interpolation='nearest', vmin=self._clim[0], vmax=self._clim[1], cmap=plt.cm.gray)

            self.slice_slider = None

        else:

            if self.array.shape[0] > 1:

                self.slice = int(round(self.array.shape[0] / 2.))

                self.slice_slider_ax = self.fig.add_axes([0.1, 0.95, 0.4, 0.03])
                self.slice_slider_ax.set_xticklabels("")
                self.slice_slider_ax.set_yticklabels("")
                self.slice_slider = Slider(self.slice_slider_ax, "3-d slice", 0, self.array.shape[0], valinit=self.slice, valfmt="%i")
                self.slice_slider.on_changed(self.update_slice)
                self.slice_slider.drawon = False

            else:

                self.slice = 0
                self.slice_slider = None

            self.image = self.ax1.imshow(self.array[self.slice, :,:], origin='lower', interpolation='nearest', vmin=self._clim[0], vmax=self._clim[1], cmap=plt.cm.gray)

        self.vmin_slider_ax = self.fig.add_axes([0.1, 0.90, 0.4, 0.03])
        self.vmin_slider_ax.set_xticklabels("")
        self.vmin_slider_ax.set_yticklabels("")
        self.vmin_slider = Slider(self.vmin_slider_ax, "vmin", self._clim[0], self._clim[1], valinit=self._clim[0])
        self.vmin_slider.on_changed(self.update_vmin)
        self.vmin_slider.drawon = False

        self.vmax_slider_ax = self.fig.add_axes([0.1, 0.85, 0.4, 0.03])
        self.vmax_slider_ax.set_xticklabels("")
        self.vmax_slider_ax.set_yticklabels("")
        self.vmax_slider = Slider(self.vmax_slider_ax, "vmax", self._clim[0], self._clim[1], valinit=self._clim[1])
        self.vmax_slider.on_changed(self.update_vmax)
        self.vmax_slider.drawon = False

        self.ax2 = self.fig.add_axes([0.6, 0.3, 0.35, 0.4])
        self.ax2.add_collection(self.lines)

        self.selected_label = self.fig.text(0.6, 0.75, "No structure selected", fontsize=18)
        x = [p.vertices[:, 0] for p in self.lines.get_paths()]
        y = [p.vertices[:, 1] for p in self.lines.get_paths()]
        xmin = np.min(x)
        xmax = np.max(x)
        ymin = np.min(y)
        ymax = np.max(y)
        self.lines.set_picker(2.)
        dx = xmax - xmin
        self.ax2.set_xlim(xmin - dx * 0.1, xmax + dx * 0.1)
        self.ax2.set_ylim(ymin * 0.5, ymax * 2.0)
        self.ax2.set_yscale('log')

        self.fig.canvas.mpl_connect('pick_event', self.line_picker)
        self.fig.canvas.mpl_connect('button_press_event', self.select_from_map)

        plt.show()

    def update_slice(self, pos=None):
        if self.array.ndim == 2:
            self.image.set_array(self.array)
        else:
            self.slice = int(round(pos))
            self.image.set_array(self.array[self.slice, :,:])

        self.remove_contour()
        self.update_contour()

        self.fig.canvas.draw()

    def update_vmin(self, vmin):
        if vmin > self._clim[1]:
            self._clim = (self._clim[1], self._clim[1])
        else:
            self._clim = (vmin, self._clim[1])
        self.image.set_clim(*self._clim)
        self.fig.canvas.draw()

    def update_vmax(self, vmax):
        if vmax < self._clim[0]:
            self._clim = (self._clim[0], self._clim[0])
        else:
            self._clim = (self._clim[0], vmax)
        self.image.set_clim(*self._clim)
        self.fig.canvas.draw()

    def select_from_map(self, event):

        # Only do this if no tools are currently selected
        if event.canvas.toolbar.mode != '':
            return

        if event.inaxes is self.ax1:

            # Find pixel co-ordinates of click
            ix = int(round(event.xdata))
            iy = int(round(event.ydata))

            if self.array.ndim == 2:
                indices = (iy, ix)
            else:
                indices = (self.slice, iy, ix)

            # Select the structure
            structure = self.dendrogram.structure_at(indices)
            self.select(structure)

            # Re-draw
            event.canvas.draw()

    def line_picker(self, event):

        # Only do this if no tools are currently selected
        if event.canvas.toolbar.mode != '':
            return

        # event.ind gives the indices of the paths that have been selected

        # Find levels of selected paths
        peaks = [event.artist.structures[i].get_peak(subtree=True)[1] for i in event.ind]

        # Find position of minimum level (may be duplicates, let Numpy decide)
        ind = event.ind[np.argmax(peaks)]

        # Extract structure
        structure = event.artist.structures[ind]

        # If 3-d, select the slice
        if self.slice_slider is not None:
            peak_index = structure.get_peak(subtree=True)
            self.slice_slider.set_val(peak_index[0][0])

        # Select the structure
        self.select(structure)

        # Re-draw
        event.canvas.draw()

    def select(self, structure):

        # Remove previously selected collection
        if self.selected_lines is not None:
            self.ax2.collections.remove(self.selected_lines)
            self.selected_lines = None

        self.remove_contour()

        if structure is None:
            self.selected_label.set_text("No structure selected")
            self.fig.canvas.draw()
            return

        self.selected = structure

        self.selected_label.set_text("Selected structure: {0}".format(structure.idx))

        # Get collection for this substructure
        self.selected_lines = self.plotter.get_lines(structure=structure)
        self.selected_lines.set_color('red')
        self.selected_lines.set_linewidth(2)
        self.selected_lines.set_alpha(0.5)

        # Add to axes
        self.ax2.add_collection(self.selected_lines)

        self.update_contour()

    def remove_contour(self):

        if self.selected_contour is not None:
            for collection in self.selected_contour.collections:
                self.ax1.collections.remove(collection)
            self.selected_contour = None

    def update_contour(self):

        if self.selected is not None:
            mask = self.selected.get_mask(subtree=True)
            if self.array.ndim == 3:
                mask = mask[self.slice, :,:]
            self.selected_contour = self.ax1.contour(mask, colors='red', linewidths=2, levels=[0.5], alpha=0.5)



# Licensed under an MIT open source license - see LICENSE

from .dendrogram import Dendrogram
from .analysis import ppv_catalog, pp_catalog
from .plot import DendrogramPlotter
from .viewer import BasicDendrogramViewer

__version__ = '0.1.0'


# Licensed under an MIT open source license - see LICENSE

import abc
import warnings

import numpy as np

from astropy.units import Quantity
from astropy.table import Table
from astropy import units as u
from astropy.wcs import WCS

from . import six
from .structure import Structure

__all__ = ['ppv_catalog', 'pp_catalog']


class MissingMetadataWarning(UserWarning):
    pass


def _qsplit(q):
    """Split a potential astropy Quantity into unit/quantity"""
    if isinstance(1. * q, Quantity):
        return q.unit, q.value

    return 1, q


def _unit(q):
    """Return the units associated with a number, array, unit, or Quantity"""
    if q is None:
        return None
    elif isinstance(1 * q, Quantity):
        return (1 * q).unit


class ScalarStatistic(object):
    #This class does all of the heavy computation
    def __init__(self, values, indices):
        """
        Compute pixel-level statistics from a scalar field, sampled at specific
        locations.

        Parameters
        ----------
        values : 1D ndarray
            data values to use
        indices: tuple of 1D arrays
            Location of each element of values. The i-th array in the tuple
            describes the ith positional dimension
        """
        self.values = values.astype(np.float)
        self.indices = indices

    def mom0(self):
        """The sum of the values"""
        return np.nansum(self.values)

    def mom1(self):
        """The intensity-weighted mean position"""
        m0 = self.mom0()
        return [np.nansum(i * self.values) / m0 for i in self.indices]

    def mom2(self):
        """The intensity-weighted covariance matrix"""
        mom1 = self.mom1()
        mom0 = self.mom0()
        v = self.values / mom0

        nd = len(self.indices)
        zyx = tuple(i - m for i, m in zip(self.indices, mom1))

        result = np.zeros((nd, nd))

        for i in range(nd):
            result[i, i] = np.nansum(v * zyx[i] ** 2)
            for j in range(i + 1, nd):
                result[i, j] = result[j, i] = np.nansum(v * zyx[i] * zyx[j])

        return result

    def mom2_along(self, direction):
        """
        Intensity-weighted variance/covariance along 1 or more directions.

        Parameters
        ----------
        direction : array like
                  One or more set of direction vectors. Need not be normalized

        Returns
        -------
        result : array
            The variance (or co-variance matrix) of the data along the
            specified direction(s).
        """
        w = np.atleast_2d(direction).astype(np.float)
        for row in w:
            row /= np.linalg.norm(row)

        result = np.dot(np.dot(w, self.mom2()), w.T)
        if result.size == 1:
            result = np.asscalar(result)
        return result

    def paxes(self):
        """
        The principal axes of the data (direction of greatest elongation)

        Returns
        -------
        result : tuple
            Ordered tuple of ndarrays

        Notes
        -----
        Each array is a normalized direction vector. The arrays
        are sorted in decreasing order of elongation of the data
        """
        mom2 = self.mom2()
        w, v = np.linalg.eig(mom2)
        order = np.argsort(w)

        return tuple(v[:, o] for o in order[::-1])

    def projected_paxes(self, axes):
        """
        The principal axes of a projection of the data onto a subspace

        Paramters
        ---------
        axes : array-like, (nnew, nold)
               The projection to take. Each row defines a unit vector in
               the new coordinate system

        Returns
        --------
        result : tuple
            Tuple of arrays (nnew items)

        Notes
        -----
        The ordered principal axes in the new space
        """
        mom2 = self.mom2_along(axes)
        w, v = np.linalg.eig(mom2)
        order = np.argsort(w)

        return tuple(v[:, o] for o in order[::-1])

    def count(self):
        """
        Number of elements in the dataset.
        """
        return self.values.size

    def surface_area(self):
        raise NotImplementedError

    def perimeter(self, plane=None):
        raise NotImplementedError


class VectorStatistic(object):
    def __init__(self, values_tuple, indices):
        raise NotImplementedError

    def divergence(self):
        raise NotImplementedError

    def curl(self):
        raise NotImplementedError


class Metadata(object):
    """
    A descriptor to wrap around metadata dictionaries.

    Lets classes reference self.x instead of self.metadata['x'],
    """

    _restrict_types = None

    def __init__(self, key, description, default=None, strict=False):
        """
        Parameters
        ----------
        key : str
               Metadata name.
        description : str
               What the quantity describes
        default : scalar
               Default value if metadata not provided
        strict : bool
               If True, raise KeyError if metadata not provided.
               This overrides default
        """
        if not isinstance(key, six.string_types):
            raise TypeError("Key is", key, type(key))
        self.key = key
        self.description = description or 'no description'
        self.default = default
        self.strict = strict

    def __get__(self, instance, type=None):

        if instance is None:
            return self

        try:
            value = instance.metadata[self.key]
        except KeyError:
            if self.strict:
                raise KeyError("Required metadata item not found: %s" % self)
            else:
                if self.default is not None:
                    warnings.warn("{0} ({1}) missing, defaulting to {2}".format(self.key, self.description, self.default),
                                  MissingMetadataWarning)
                value = self.default

        if value is not None and self._restrict_types is not None:
            if isinstance(value, self._restrict_types):
                return value
            else:
                raise TypeError("{0} should be an instance of {1}".format(self.key, ' or '.join([x.__name__ for x in self._restrict_types])))
        else:
            return value

    def __str__(self):
        return "%s (%s)" % (self.key, self.description)


class MetadataQuantity(Metadata):
    _restrict_types = (u.UnitBase, u.Quantity)


class MetadataWCS(Metadata):
    _restrict_types = (WCS,)


class SpatialBase(object):

    __metaclass__ = abc.ABCMeta

    wavelength = MetadataQuantity('wavelength', 'Wavelength')
    spatial_scale = MetadataQuantity('spatial_scale', 'Pixel width/height')
    beam_major = MetadataQuantity('beam_major', 'Major FWHM of beam')
    beam_minor = MetadataQuantity('beam_minor', 'Minor FWHM of beam')
    data_unit = MetadataQuantity('data_unit', 'Units of the pixel values', strict=True)
    wcs = MetadataWCS('wcs', 'WCS object')

    @abc.abstractmethod
    def _sky_paxes(self):
        raise NotImplementedError()

    def _world_pos(self):
        xyz = self.stat.mom1()[::-1]
        if self.wcs is None:
            return xyz[::-1] * u.pixel
        else:
            # TODO: set units correctly following WCS
            # We use origin=0 since the indices come from Numpy indexing
            return self.wcs.all_pix2world([xyz], 0).ravel()[::-1]

    @abc.abstractproperty
    def flux(self):
        raise NotImplementedError

    @abc.abstractproperty
    def x_cen(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def y_cen(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def position_angle(self):
        raise NotImplementedError()

    @property
    def major_sigma(self):
        """
        Major axis of the projection onto the position-position (PP) plane,
        computed from the intensity weighted second moment in direction of
        greatest elongation in the PP plane.
        """
        dx = self.spatial_scale or u.pixel
        a, b = self._sky_paxes()
        # We need to multiply the second moment by two to get the major axis
        # rather than the half-major axis.
        return dx * np.sqrt(self.stat.mom2_along(a))

    @property
    def minor_sigma(self):
        """
        Minor axis of the projection onto the position-position (PP) plane,
        computed from the intensity weighted second moment perpendicular to
        the major axis in the PP plane.
        """
        dx = self.spatial_scale or u.pixel
        a, b = self._sky_paxes()
        # We need to multiply the second moment by two to get the minor axis
        # rather than the half-minor axis.
        return dx * np.sqrt(self.stat.mom2_along(b))

    @property
    def radius(self):
        """
        Geometric mean of ``major_sigma`` and ``minor_sigma``.
        """
        u, a = _qsplit(self.major_sigma)
        u, b = _qsplit(self.minor_sigma)
        return u * np.sqrt(a * b)

    @property
    def area_ellipse(self):
        """
        The area of the ellipse defined by the second moments, where the
        semi-major and semi-minor axes used are the HWHM (half-width at
        half-maximum) derived from the moments.
        """
        return np.pi * self.major_sigma * self.minor_sigma * (2.3548 * 0.5) ** 2

    def to_mpl_ellipse(self, **kwargs):
        """
        Returns a Matplotlib ellipse representing the first and second moments
        of the structure.

        Any keyword arguments are passed to :class:`~matplotlib.patches.Ellipse`
        """
        from matplotlib.patches import Ellipse
        return Ellipse((self.x_cen.value, self.y_cen.value),
                        self.major_sigma.value * 2.3548,
                        self.minor_sigma.value * 2.3548,
                        angle=self.position_angle.value,
                        **kwargs)

class PPVStatistic(SpatialBase):
    """
    Compute properties of structures in a position-position-velocity (PPV)
    cube.

    Parameters
    ----------
    structure : `~astrodendro.structure.Structure` instance
        The structure to compute the statistics for
    metadata : dict
         Key-value pairs of metadata
    """

    velocity_scale = MetadataQuantity('velocity_scale', 'Velocity channel width')
    vaxis = Metadata('vaxis', 'Index of velocity axis (numpy convention)', default=0)

    def __init__(self, stat, metadata=None):
        if isinstance(stat, Structure):
            self.stat = ScalarStatistic(stat.values(subtree=True),
                                        stat.indices(subtree=True))
        else:
            self.stat = stat
        if len(self.stat.indices) != 3:
            raise ValueError("PPVStatistic can only be used on 3-d datasets")
        self.metadata = metadata or {}

    def _sky_paxes(self):
        vaxis = self.vaxis
        ax = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        ax.pop(vaxis)
        a, b = self.stat.projected_paxes(ax)
        a = list(a)
        a.insert(0, vaxis)
        b = list(b)
        b.insert(0, vaxis)
        return a, b

    @property
    def x_cen(self):
        """
        The mean position of the structure in the x direction.
        """
        p = self._world_pos()
        return p[2] if self.vaxis != 2 else p[1]

    @property
    def y_cen(self):
        """
        The mean position of the structure in the y direction.
        """
        p = self._world_pos()
        return p[1] if self.vaxis == 0 else p[0]

    @property
    def v_cen(self):
        """
        The mean velocity of the structure (where the velocity axis can be
        specified by the ``vaxis`` metadata parameter, which defaults to 0
        following the Numpy convention - the third axis in the FITS convention).
        """
        p = self._world_pos()
        return p[self.vaxis]

    @property
    def flux(self):
        """
        The integrated flux of the structure, in Jy (note that this does not
        include any kind of background subtraction, and is just a plain sum of
        the values in the structure, converted to Jy).
        """
        from .flux import compute_flux
        return compute_flux(self.stat.mom0() * self.data_unit,
                            u.Jy,
                            wavelength=self.wavelength,
                            spatial_scale=self.spatial_scale,
                            velocity_scale=self.velocity_scale,
                            beam_major=self.beam_major,
                            beam_minor=self.beam_minor)

    @property
    def v_rms(self):
        """
        Intensity-weighted second moment of velocity (where the velocity axis
        can be specified by the ``vaxis`` metadata parameter, which defaults to
        0 following the Numpy convention - the third axis in the FITS
        convention).
        """
        dv = self.velocity_scale or u.pixel
        ax = [0, 0, 0]
        ax[self.vaxis] = 1
        return dv * np.sqrt(self.stat.mom2_along(ax))

    @property
    def position_angle(self):
        """
        The position angle of sky_maj, sky_min in degrees counter-clockwise
        from the +x axis (note that this is the +x axis in pixel coordinates,
        which is the ``-x`` axis for conventional astronomy images).
        """
        a, b = self._sky_paxes()
        a.pop(self.vaxis)
        return np.degrees(np.arctan2(a[0], a[1])) * u.degree

    @property
    def area_exact(self):
        """
        The exact area of the structure on the sky.
        """
        dx = self.spatial_scale or u.pixel
        indices = zip(*tuple(self.stat.indices[i] for i in range(3) if i != self.vaxis))
        return len(set(indices)) * dx**2


class PPStatistic(SpatialBase):
    """
    Compute properties of structures in a position-position (PP) cube.

    Parameters
    ----------
    structure : `~astrodendro.structure.Structure` instance
        The structure to compute the statistics for
    metadata : dict
         Key-value pairs of metadata
    """

    def __init__(self, stat, metadata=None):
        if isinstance(stat, Structure):
            self.stat = ScalarStatistic(stat.values(subtree=True),
                                        stat.indices(subtree=True))
        else:
            self.stat = stat
        if len(self.stat.indices) != 2:
            raise ValueError("PPStatistic can only be used on 2-d datasets")
        self.metadata = metadata or {}

    def _sky_paxes(self):
        return self.stat.paxes()

    @property
    def flux(self):
        """
        The integrated flux of the structure, in Jy (note that this does not
        include any kind of background subtraction, and is just a plain sum of
        the values in the structure, converted to Jy).
        """
        from .flux import compute_flux
        return compute_flux(self.stat.mom0() * self.data_unit,
                            u.Jy,
                            wavelength=self.wavelength,
                            spatial_scale=self.spatial_scale,
                            beam_major=self.beam_major,
                            beam_minor=self.beam_minor)

    @property
    def position_angle(self):
        """
        The position angle of sky_maj, sky_min in degrees counter-clockwise
        from the +x axis.
        """
        a, b = self._sky_paxes()
        return np.degrees(np.arctan2(a[0], a[1])) * u.degree

    @property
    def x_cen(self):
        """
        The mean position of the structure in the x direction (in pixel
        coordinates, or in world coordinates if the WCS transformation is
        available in the meta-data).
        """
        return self._world_pos()[1]

    @property
    def y_cen(self):
        """
        The mean position of the structure in the y direction (in pixel
        coordinates, or in world coordinates if the WCS transformation is
        available in the meta-data).
        """
        return self._world_pos()[0]

    @property
    def area_exact(self):
        """
        The exact area of the structure on the sky.
        """
        dx = self.spatial_scale or u.pixel
        return self.stat.count() * dx**2


class PPPStatistic(object):

    def __init__(self, rhostat, vstat, metadata=None):
        """
        Derive properties from PPP density and velocity fields.

        This is not currently implemented

        Parameters
        ----------
        rhostat : ScalarStatistic instance
        vstat : VectorStatistic instance
        """
        raise NotImplementedError()

    @property
    def mass(self):
        pass

    @property
    def volume(self):
        pass

    @property
    def surface_area(self):
        pass

    @property
    def virial(self):
        pass

    @property
    def v_rms(self):
        pass

    @property
    def vz_rms(self):
        pass

    @property
    def pressure_vz(self):
        pass

    @property
    def pressure(self):
        pass


def _make_catalog(structures, fields, metadata, statistic):
    """
    Make a catalog from a list of structures
    """

    result = None

    for struct in structures:
        stat = ScalarStatistic(struct.values(subtree=True),
                               struct.indices(subtree=True))
        stat = statistic(stat, metadata)
        row = {}
        for lbl in fields:
            row[lbl] = getattr(stat, lbl)

        row = dict((lbl, getattr(stat, lbl))
                   for lbl in fields)
        row.update(_idx=struct.idx)

        # first row
        if result is None:
            sorted_row_keys = sorted(row.keys())
            try:
                result = Table(names=sorted_row_keys,
                               dtype=[int if x == '_idx' else float for x in sorted_row_keys])
            except TypeError:  # dtype was called dtypes in older versions of Astropy
                result = Table(names=sorted_row_keys,
                               dtypes=[int if x == '_idx' else float for x in sorted_row_keys])            
            for k, v in row.items():
                try:  # Astropy API change
                    result[k].unit = _unit(v)
                except AttributeError:
                    result[k].units = _unit(v)

        # astropy.table.Table should in future support setting row items from
        # quantities, but for now we need to strip off the quantities
        new_row = {}
        for x in row:
            if row[x] is not None:  # in Astropy 0.3+ we no longer need to exclude None items
                if isinstance(row[x], Quantity):
                    new_row[x] = row[x].value
                else:
                    new_row[x] = row[x]
        result.add_row(new_row)

    result.sort('_idx')

    return result


def ppv_catalog(structures, metadata, fields=None, verbose=True):
    """
    Iterate over a collection of position-position-velocity (PPV) structures,
    extracting several quantities from each, and building a catalog.

    Parameters
    ----------
    structures : iterable of Structures
         The structures to catalog (e.g., a dendrogram)
    metadata : dict
        The metadata used to compute the catalog
    fields : list of strings, optional
        The quantities to extract. If not provided,
        defaults to all PPV statistics
    verbose : bool, optional
        If True (the default), will generate warnings
        about missing metadata

    Returns
    -------
    table : a :class:`~astropy.table.table.Table` instance
        The resulting catalog
    """
    fields = fields or ['major_sigma', 'minor_sigma', 'radius', 'area_ellipse', 'area_exact',
                        'position_angle', 'v_rms', 'x_cen', 'y_cen', 'v_cen', 'flux']
    with warnings.catch_warnings():
        warnings.simplefilter("once" if verbose else 'ignore', category=MissingMetadataWarning)
        return _make_catalog(structures, fields, metadata, PPVStatistic)


def pp_catalog(structures, metadata, fields=None, verbose=True):
    """
    Iterate over a collection of position-position (PP) structures, extracting
    several quantities from each, and building a catalog.

    Parameters
    ----------
    structures : iterable of Structures
         The structures to catalog (e.g., a dendrogram)
    metadata : dict
        The metadata used to compute the catalog
    fields : list of strings, optional
        The quantities to extract. If not provided,
        defaults to all PPV statistics
    verbose : bool, optional
        If True (the default), will generate warnings
        about missing metadata

    Returns
    -------
    table : a :class:`~astropy.table.table.Table` instance
        The resulting catalog
    """
    fields = fields or ['major_sigma', 'minor_sigma', 'radius', 'area_ellipse', 'area_exact',
                        'position_angle', 'x_cen', 'y_cen', 'flux']
    with warnings.catch_warnings():
        warnings.simplefilter("once" if verbose else 'ignore', category=MissingMetadataWarning)
        return _make_catalog(structures, fields, metadata, PPStatistic)



# Licensed under an MIT open source license - see LICENSE

from matplotlib.collections import LineCollection


class StructureCollection(LineCollection):

    @property
    def structures(self):
        return self._structures

    @structures.setter
    def structures(self, values):
        self._structures = values



# Licensed under an MIT open source license - see LICENSE

import os

import numpy as np

from .util import parse_dendrogram
from .handler import IOHandler

# Import and export

# FITS file signature as per RFC 4047
FITS_SIGNATURE = (b"\x53\x49\x4d\x50\x4c\x45\x20\x20\x3d\x20\x20\x20\x20\x20"
                  b"\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20\x20"
                  b"\x20\x54")


def is_fits(filename, mode='r'):
    if mode == 'r' and os.path.exists(filename):
        fileobj = open(filename, 'rb')
        sig = fileobj.read(30)
        return sig == FITS_SIGNATURE
    elif filename.lower().endswith(('.fits', '.fits.gz', '.fit', '.fit.gz')):
        return True
    else:
        return False


def dendro_export_fits(d, filename):
    """Export the dendrogram 'd' to the FITS file 'filename'"""
    from astropy.io import fits

    hdus = [fits.PrimaryHDU(),
            fits.ImageHDU(d.data),
            fits.ImageHDU(d.index_map),
            fits.ImageHDU(np.array([ord(x) for x in  d.to_newick()]))]
    hdulist = fits.HDUList(hdus)

    hdulist.writeto(filename, clobber=True)


def dendro_import_fits(filename):
    """Import 'filename' and construct a dendrogram from it"""
    from astropy.io import fits

    with fits.open(filename) as hdus:
        data = hdus[1].data
        index_map = hdus[2].data
        newick = ''.join(chr(x) for x in hdus[3].data.flat)

    return parse_dendrogram(newick, data, index_map)


FITSHandler = IOHandler(identify=is_fits,
                        export_dendro=dendro_export_fits,
                        import_dendro=dendro_import_fits)



# Licensed under an MIT open source license - see LICENSE

from .fits import FITSHandler
from .hdf5 import HDF5Handler


IO_FORMATS = {
    'fits': FITSHandler,
    'hdf5': HDF5Handler
}


def _valid(formats):
    return " and ".join(["'{0}'".format(key) for key in formats])


def load_dendrogram(filename, format=None):
    if format is not None:
        return IO_FORMATS[format].import_dendro(filename)
    else:
        for handler in IO_FORMATS.values():
            if handler.identify(filename, mode='r'):
                return handler.import_dendro(filename)
    raise IOError("Could not automatically identify file format - use the "
                  "format= option to specify which format to use (valid "
                  "options are {0})".format(_valid(IO_FORMATS)))


def save_dendrogram(dendrogram, filename, format=None):
    if format is not None:
        return IO_FORMATS[format].export_dendro(dendrogram, filename)
    else:
        for handler in IO_FORMATS.values():
            if handler.identify(filename, mode='w'):
                return handler.export_dendro(dendrogram, filename)
    raise IOError("Could not automatically identify file format - use the "
                  "format= option to specify which format to use (valid "
                  "options are {0})".format(_valid(IO_FORMATS)))



import pytest
import numpy as np

from astropy import units as u

from ..flux import compute_flux


def strip_parentheses(string):
    return string.replace('(', '').replace(')', '')

COMBINATIONS = \
    [
        (np.array([1, 2, 3]) * u.Jy, u.Jy, {}, 6 * u.Jy),
(np.array([1, 2, 3]) * u.mJy, u.Jy, {}, 0.006 * u.Jy),
(np.array([1, 2, 3]) * u.erg / u.cm ** 2 / u.s / u.Hz, u.Jy, {}, 6e23 * u.Jy),
(np.array([1, 2, 3]) * u.erg / u.cm ** 2 / u.s / u.micron, u.Jy, {'wavelength': 2 * u.micron}, 8005538284.75565 * u.Jy),
(np.array([1, 2, 3]) * u.Jy / u.arcsec ** 2, u.Jy, {'spatial_scale': 2 * u.arcsec}, 24. * u.Jy),
(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, {'spatial_scale': 2 * u.arcsec, 'beam_major': 1 * u.arcsec, 'beam_minor': 0.5 * u.arcsec}, 42.36166269526079 * u.Jy),
    ]


@pytest.mark.parametrize(('input_quantities', 'output_unit', 'keywords', 'output'), COMBINATIONS)
def test_compute_flux(input_quantities, output_unit, keywords, output):
    q = compute_flux(input_quantities, output_unit, **keywords)
    np.testing.assert_allclose(q.value, output.value)
    assert q.unit == output.unit


def test_monochromatic_wav_missing():
    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.erg / u.cm ** 2 / u.s / u.micron, u.Jy)
    assert exc.value.args[0] == 'wavelength is needed to convert from erg / (cm2 micron s) to Jy'


def test_monochromatic_wav_invalid_units():
    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.erg / u.cm ** 2 / u.s / u.micron, u.Jy, wavelength=3 * u.L)
    assert exc.value.args[0] == 'wavelength should be a physical length'


def test_surface_brightness_scale_missing():
    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.arcsec ** 2, u.Jy)
    assert strip_parentheses(exc.value.args[0]) == 'spatial_scale is needed to convert from Jy / arcsec2 to Jy'


def test_surface_brightness_invalid_units():
    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.arcsec ** 2, u.Jy, spatial_scale=3 * u.m)
    assert exc.value.args[0] == 'spatial_scale should be an angle'


def test_per_beam_scale_missing():

    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, beam_major=3 * u.arcsec, beam_minor=2. * u.arcsec)
    assert strip_parentheses(exc.value.args[0]) == 'spatial_scale is needed to convert from Jy / beam to Jy'

    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, spatial_scale=3 * u.arcsec, beam_minor=2. * u.arcsec)
    assert strip_parentheses(exc.value.args[0]) == 'beam_major is needed to convert from Jy / beam to Jy'

    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, spatial_scale=3 * u.arcsec, beam_major=2. * u.arcsec)
    assert strip_parentheses(exc.value.args[0]) == 'beam_minor is needed to convert from Jy / beam to Jy'


def test_per_beam_invalid_units():

    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, spatial_scale=3 * u.arcsec, beam_major=3 * u.m, beam_minor=2. * u.arcsec)
    assert exc.value.args[0] == 'beam_major should be an angle'

    with pytest.raises(ValueError) as exc:
        compute_flux(np.array([1, 2, 3]) * u.Jy / u.beam, u.Jy, spatial_scale=3 * u.arcsec, beam_major=3 * u.arcsec, beam_minor=2. * u.m)
    assert exc.value.args[0] == 'beam_minor should be an angle'



# Licensed under an MIT open source license - see LICENSE

import sys

import numpy as np
import matplotlib.pyplot as plt
plt.switch_backend('Agg')

from .. import Dendrogram


class TestRecursionLimit(object):
    """
    Test that we can efficiently compute deep dendrogram trees
    without hitting the recursion limit.
    Note: plot() uses recursion but we should be able to *compute*
    dendrograms without using deep recursion, even if we aren't
    yet able to plot them without using recursion.
    """
    def setup_method(self, method):
        self._oldlimit = sys.getrecursionlimit()
        sys.setrecursionlimit(100)  # Reduce recursion limit dramatically (default is 1000)
        size = 10000  # number of leaves desired in the dendrogram
        self._make_data(size)

    def _make_data(self, size):
        data1 = np.arange(size * 2)  # first row
        data2 = np.arange(size * 2)  # second row
        data2[::2] += 2
        data1[-1] = 0  # set the last pixel in the first row to zero, to trigger a deep ancestor search
        data = np.vstack((data1, data2))
        self.data = data
        self.size = size

        # result looks like this:
        # [[ 0, 1, 2, 3, 4, 5, ...],
        #  [ 2, 1, 4, 3, 6, 5, ...]]
        # Notice every second column has a local maximum
        # so there are [size] local maxima in the array

    def test_compute(self):
        d = Dendrogram.compute(self.data)
        assert len(d.leaves) == self.size, "We expect {n} leaves, not {a}.".format(n=self.size, a=len(d.leaves))

    def test_computing_level(self):
        d = Dendrogram.compute(self.data)

        # Now pick a structure near the middle of the dendrogram:
        mid_structure = d.structure_at((0, self.size // 2))

        # Compute its level:
        sys.setrecursionlimit(100000)
        _ = mid_structure.level

        # Check that .level satisfies the recurrence relation:
        # 0 if root else parent.level + 1
        for structure in d.all_structures:
            if structure.parent is None:
                assert structure.level == 0
            else:
                assert structure.level == structure.parent.level + 1


    def test_plot(self):
        sys.setrecursionlimit(self._oldlimit)
        ax = plt.gca()
        sys.setrecursionlimit(100)

        d = Dendrogram.compute(self.data)
        p = d.plotter()
        p.plot_tree(ax)

    def teardown_method(self, method):
        sys.setrecursionlimit(self._oldlimit)



#!/usr/bin/env python

from setuptools import setup, Command

from distutils.command.build_py import build_py


class DendroTest(Command):

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):

        import os
        import shutil
        import tempfile

        # First ensure that we build the package so that 2to3 gets executed
        self.reinitialize_command('build', inplace=False)
        self.run_command('build')
        build_cmd = self.get_finalized_command('build')
        new_path = os.path.abspath(build_cmd.build_lib)

        # Copy the build to a temporary directory for the purposes of testing
        # - this avoids creating pyc and __pycache__ directories inside the
        # build directory
        tmp_dir = tempfile.mkdtemp(prefix='astrodendro-test-')
        testing_path = os.path.join(tmp_dir, os.path.basename(new_path))
        shutil.copytree(new_path, testing_path)

        import sys
        import subprocess

        errno = subprocess.call([sys.executable, os.path.abspath('runtests.py')], cwd=testing_path)
        raise SystemExit(errno)


setup(name='astrodendro',
      version='0.1.0',
      url='http://www.dendrograms.org',
      description='Python package for computation of astronomical dendrograms',
      author='Thomas Robitaille, Chris Beaumont, Adam Ginsburg, Braden MacDonald, and Erik Rosolowsky',
      author_email='thomas.robitaille@gmail.com',
      packages=['astrodendro', 'astrodendro.io', 'astrodendro.tests'],
      package_data={'astrodendro.tests':['*.npz', 'benchmark_data/*fits']},
      provides=['astrodendro'],
      requires=['numpy'],
      cmdclass={'build_py': build_py, 'test': DendroTest},
      keywords=['Scientific/Engineering'],
      classifiers=[
                   "Development Status :: 4 - Beta",
                   "Programming Language :: Python",
                   "License :: OSI Approved :: MIT License",
                  ],
     )



"""
Realistic data for unit tests and benchmarks
This module provides a 107x107x602 pixel data cube where all the flux values
greater than 1.4 represent real data from L1448 13co, but all other (lesser)
values are randomly generated.
"""
import os
try:
    import gzip
except AttributeError:
    # gzip tries to import from the "io" package, which causes an error
    # if the user's current directory is the "astrodendro" directory
    # which has its own "io" package.
    raise RuntimeError("This test cannot be run from a folder with an 'io'"
                       " subfolder. Please change to a different directory"
                       " and re-run this test.")

import numpy as np

# First, load the data cube from the file:
_datafile_path = os.path.join(os.path.dirname(__file__), 'sample-data-hl.npz')

# data file contains saved flux_values and indices from L1448 13co,
# a data cube which originally had a shape of (107, 107, 602)
# but filtered to only include data points above value of 1.4

arrays = np.load(_datafile_path)
_flux_values = arrays['flux_values']
_indices = arrays['coords']

# Create a new data cube filled with random values no greater than 1.4
# (these will later be filtered out, but that filtering should be part of
# the tests and benchmark)

data = np.random.normal(0, 0.25, (107, 107, 602))

for i in range(_flux_values.size):
    data[tuple(_indices[i])] = _flux_values[i]



# Licensed under an MIT open source license - see LICENSE

import numpy as np


def cached_property(func):
    memo = {}
    def result(self):
        if self not in memo:
            memo[self] = func(self)
        return memo[self]

    return property(result)


def prefix_visit(s, key=None, reverse=False):
    todo = [s]
    while todo:
        st = todo.pop(0)
        yield st
        children = st.children
        if key is not None:
            children = sorted(children, key=key, reverse=reverse)
        todo = children + todo


class Structure(object):
    """
    A structure in the dendrogram, for example a leaf or a branch.

    A structure that is part of a dendrogram knows which other structures it is
    related to. For example, it is possible to get the parent structure
    containing the present structure ``s`` by using the ``parent`` attribute::

        >>> s.parent
        <Structure type=branch idx=2152>

    Likewise, the ``children`` attribute can be used to get a list of all
    sub-structures::

        >>> s.children
        [<Structure type=branch idx=1680>, <Structure type=branch idx=5771>]

    A number of attributes and methods are available to explore the structure
    in more detail, such as the ``indices`` and ``values`` methods, which
    return the indices and values of the pixels that are part of the
    structure. These and other methods have a ``subtree=`` option, which if
    ``True`` (the default) returns the quantities related to structure and all
    sub-structures, and if ``False`` includes only the pixels that are part of
    the structure, but excluding any sub-structure.
    """

    ###########################################################################
    #   The following methods are used only by the Dendrogram class,          #
    #   for computing the dendrogram and should never be used manually:       #
    ###########################################################################

    def __init__(self, indices, values, children=[], idx=None, dendrogram=None):

        self._dendrogram = dendrogram
        self.parent = None
        self.children = children

        # Make sure that the children have a reference to the present structure
        for child in children:
            child.parent = self

        if np.isscalar(values):  # values are for a single pixel
            self._indices = [indices]
            self._values = [values]
            self._vmin, self._vmax = values, values
        elif isinstance(indices, list) and isinstance(values, list):
            self._indices = indices
            self._values = values
            self._vmin, self._vmax = min(values), max(values)
        else:  # could be an array or iterator
            self._indices = [x for x in indices]
            self._values = [x for x in values]
            self._vmin, self._vmax = min(values), max(values)

        self._smallest_index = min(self._indices)

        self.idx = idx

        self._reset_cache()

    def _reset_cache(self):
        self._level = None
        self._ancestor = None
        self._descendants = None
        self._npix_total = None
        self._peak = None
        self._peak_subtree = None
        self._tree_index = None

    @property
    def smallest_index(self):
        return self._smallest_index

    @smallest_index.setter
    def smallest_index(self, value):
        self._smallest_index = value

    @property
    def parent(self):
        """
        The parent structure containing the present structure.
        """
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def children(self):
        """
        A list of all the sub-structures contained in the present structure.
        """
        return self._children

    @children.setter
    def children(self, value):
        self._children = value

    @property
    def is_leaf(self):
        """
        Whether the present structure is a leaf.
        """
        return not self.children

    @property
    def is_branch(self):
        """
        Whether the present structure is a branch.
        """
        return not self.is_leaf

    def indices(self, subtree=True):
        """
        The indices of the pixels in this branch.

        Parameters
        ----------
        subtree : bool, optional
            Whether to recursively include all sub-structures
        """

        if self._tree_index is not None:
            return self._tree_index.indices(self.idx, subtree=subtree)

        if subtree:
            sub_indices = [self.indices(subtree=False)]
            for child in self.children:
                sub_indices.append(child.indices(subtree=True))
            return tuple(np.hstack(arrs) for arrs in zip(*(sub_indices)))
        else:
            return tuple(np.atleast_1d(i) for i in zip(*self._indices))

    def values(self, subtree=True):
        """
        The values of the pixels in this branch.

        Parameters
        ----------
        subtree : bool, optional
            Whether to recursively include all sub-structures
        """

        if self._tree_index is not None:
            return self._tree_index.values(self.idx, subtree=subtree)

        if subtree:
            sub_values = [self.values(subtree=False)]
            for child in self.children:
                sub_values.append(child.values(subtree=True))
            return np.hstack(sub_values)
        else:
            return np.atleast_1d(self._values)

    @property
    def vmin(self):
        """
        The minimum value of pixels belonging to the branch (excluding sub-structure).
        """
        return self._vmin

    @property
    def vmax(self):
        """
        The maximum value of pixels belonging to the branch (excluding sub-structure).
        """
        return self._vmax

    def _add_pixel(self, index, value):
        """
        Add a pixel to this leaf.

        This is a private method only intended for use by
        :meth:`~astrodendro.dendrogram.Dendrogram.compute`

        Parameters
        ----------
        index : tuple
            The pixel coordinates
        value : float
            The value of the pixel
        """
        self._indices.append(index)
        self._values.append(value)
        self._vmin, self._vmax = min(value, self.vmin), max(value, self.vmax)
        self._smallest_index = min(self._smallest_index, index)
        self._reset_cache()

    def _merge(self, structure):
        """
        Merge a structure with the present structure.

        This is a private method only intended for use by
        :meth:`~astrodendro.dendrogram.Dendrogram.compute`
        """
        self._indices.extend(structure._indices)
        self._values.extend(structure._values)
        self._vmin, self._vmax = min(structure.vmin, self.vmin), max(structure.vmax, self.vmax)
        self._smallest_index = min(structure._smallest_index, self._smallest_index)
        self._reset_cache()

    ###########################################################################
    #   The following methods can be used during OR after computation         #
    ###########################################################################

    @property
    def height(self):
        """
        This is defined as the minimum value in the children structures, or the
        peak value of the present structure if it has no children.
        """
        return min(c.vmin for c in self.children) if self.children else self.vmax

    @property
    def ancestor(self):
        """
        Find the ancestor of this leaf/branch non-recursively.

        This always returns an object (may return self if the object has no
        parent). Results are partially cached to reduce recursion depth. The
        caching assumes that once an object has been given a parent, that
        parent will never change.

        This method should be equivalent to:

            if self.parent == None:
                return self
            else:
                return self.parent.ancestor
        """

        if self.parent is None:
            return self

        if not self._ancestor:
            self._ancestor = self.parent

        # Use a loop rather than recursion to update the cached ancestor:
        while self._ancestor.parent:
            a = self._ancestor
            if a._ancestor:
                self._ancestor = a._ancestor
            else:
                self._ancestor = a.parent

        return self._ancestor

    ###########################################################################
    #   The following methods are only reliable after the entire tree is      #
    #   computed. They should not be used in dendrogram.py                    #
    ###########################################################################

    def _fill_footprint(self, array, level, recursive=True):
        """
        Set all corresponding points in `array` to the level of the structure.

        Parameters
        ----------
        array : `~numpy.ndarray`
            The array to write the footprint to
        recursive : bool
            Whether to also add the footprint for the sub-structures
        """
        if recursive:
            for child in self.children:
                child._fill_footprint(array, level + 1)
        array[self.indices(subtree=False)] = level

    @property
    def level(self):
        """
        The level of the structure, i.e. how many structures need to be traversed to reach the present structure.

        This is 0 for structures in the trunk, with values increasing in steps of 1
        towards the leaves.
        """

        if self._level is None:
            if not self.parent:
                self._level = 0
            elif self.parent._level is not None:
                self._level = self.parent._level + 1
            else:
                # We could just use:
                #  self._level = self.parent.level + 1
                # But to avoid recursion, and keep things fast, we do it this way instead:
                obj = self.parent
                diff = 1
                while obj._level is None:
                    obj = obj.parent
                    diff += 1
                    # Note: we are counting on the dendrogram computation to
                    # ensure that ._level=0 for all structures in the trunk
                self._level = obj._level + diff
                self.parent._level = self._level - 1

        return self._level

    @cached_property
    def newick(self):
        """
        Newick representation of this structure.
        """
        if self.idx is None:
            raise ValueError("Cannot return Newick representation if idx is not set")
        if self.children:
            newick_items = [child.newick for child in self.children]
            return "(%s)%s:%.3f" % (','.join(newick_items), self.idx, self.height)
        else:
            return "%i:%.3f" % (self.idx, self.height)

    @property
    def descendants(self):
        """
        Get a flattened list of all child leaves and branches.
        """

        if self._descendants is None:
            self._descendants = []
            to_add = [self]  # branches with children we will need to add to the list
            while True:
                children = []
                list(map(children.extend, [branch.children for branch in to_add]))
                self._descendants.extend(children)
                # Then proceed, essentially recursing through child branches:
                to_add = [b for b in children if not b.is_leaf]
                if not to_add:
                    break

        return self._descendants

    def get_npix(self, subtree=True):
        """
        Return the number of pixels in this structure.

        Parameters
        ----------
        subtree : bool, optional
            Whether to recursively include all sub-structures when counting the
            pixels.

        Returns
        -------
        n_pix : int
            The number of pixels in this structure
        """

        if subtree:
            if self._npix_total is None:
                self._npix_total = self.values(subtree=True).size
            return self._npix_total
        else:
            return len(self.values(subtree=False))

    def get_peak(self, subtree=True):
        """
        Return (index, value) for the pixel with maximum value.

        Parameters
        ----------
        subtree : bool, optional
            Whether to recursively include all sub-structures when searching
            for the peak.

        Returns
        -------
        index : tuple
            The n-dimensional index of the peak pixel
        value : float
            The value of the peak pixel
        """
        # populate caches without recursion
        def key(x):
            return x[1]

        if self._peak is None:
            for s in reversed(list(prefix_visit(self))):
                s._peak = (s._indices[s._values.index(s.vmax)],
                          s.vmax)
                if s.is_leaf:
                    s._peak_subtree = s._peak
                else:
                    s._peak_subtree = max((c._peak_subtree
                                           for c in s.children),
                                           key=key)
                    s._peak_subtree = max(s._peak_subtree, s._peak, key=key)

        if not subtree:
            return self._peak

        return self._peak_subtree

    def __repr__(self):
        if self.is_leaf:
            return "<Structure type=leaf idx={0}>".format(self.idx)
        else:
            return "<Structure type=branch idx={0}>".format(self.idx)

    def sorted_leaves(self, sort_key=lambda s: s.get_peak(subtree=True)[1],
                          reverse=False, subtree=True):
        """
        Return a list of sorted leaves.

        Parameters
        ----------
        sort_key : function, optional
            A function which given a structure will return a scalar that is
            then used for sorting. By default, this is set to a function that
            returns the peak value of a structure (including descendants).
        reverse : bool, optional
            Whether to reverse the sorting.
        subtree : bool, optional
            Whether to recursively include all sub-structures in the list.

        Returns
        -------
        leaves : list
            A list of sorted leaves
        """
        if self.is_leaf:
            return [self]
        if not subtree:
            return [s for s in sorted(self.children, key=sort_key,
                                      reverse=reverse)
                    if s.is_leaf]
        # flip reverse keyword, so that nodes are properly
        # sorted after reversed()
        sts = reversed(list(prefix_visit(self, key=sort_key,
                                         reverse=not reverse)))
        return [s for s in sts if s.is_leaf]

    def get_mask(self, shape=None, subtree=True):
        """
        Return a boolean mask outlining the structure.

        Parameters
        ----------
        shape : tuple, optional
            The shape of the array upon which to compute the mask. This is only
            required if the structure is not attached to a dendrogram.
        subtree : bool, optional
            Whether to recursively include all sub-structures in the mask.

        Returns
        -------
        mask : `~numpy.ndarray`
            The mask outlining the structure (``False`` values are used outside
            the structure, and ``True`` values inside).
        """
        if shape is None:
            shape = self._dendrogram.data.shape
        indices = self.indices(subtree=True) if subtree else self.indices
        mask = np.zeros(shape, dtype=bool)
        mask[indices] = True
        return mask



# Licensed under an MIT open source license - see LICENSE

import numpy as np
from ..dendrogram import Dendrogram, TreeIndex

def assert_permuted_fancyindex(x, y):
    """ Assert that two fancy indices (tuples of integer ndarrays)
    are permutations of each other
    """
    if not isinstance(x, tuple) or not(isinstance(x[0], np.ndarray)):
        raise TypeError("First argument not a fancy index: %s" % x)

    if not isinstance(y, tuple) or not(isinstance(y[0], np.ndarray)):
        raise TypeError("Second argument not a fancy index: %s" % y)

    dtype = [('%i' % i, 'i') for i in range(len(x))]
    x = np.array(list(zip(*x)), dtype=dtype)
    y = np.array(list(zip(*y)), dtype=dtype)
    np.testing.assert_array_equal(np.sort(x),
                                  np.sort(y))

def assert_identical_fancyindex(x, y):
    for xx, yy in zip(x, y):
        np.testing.assert_array_equal(xx, yy)


class TestIndex(object):

    def setup_method(self, method):
        pass

    def assert_valid_index(self, d, index):
        """Assert that a dendrogram index is correct"""

        #subtree=False is a permutation of np.where(index_map == x)
        for s in d.all_structures:
            ind = index.indices(s.idx, subtree=False)
            expected = np.where(d.index_map == s.idx)
            assert_permuted_fancyindex(ind, expected)

        #subtree=True is the same, but includes descendents
        for s in d.all_structures:
            ind = index.indices(s.idx, subtree=True)
            expected = [np.where(d.index_map == ss.idx) for
                        ss in [s] + s.descendants]
            expected = tuple(np.hstack(i) for i in zip(*expected))
            assert_permuted_fancyindex(ind, expected)

    def test_single_trunk(self):
        data = np.array([[1, 1, 1, 1],
                         [1, 5, 1, 1],
                         [1, 4, 3, 5],
                         [1, 1, 1, 4]])
        d = Dendrogram.compute(data)
        assert len(d) == 3
        self.assert_valid_index(d, TreeIndex(d))

    def test_two_trunk(self):
        data = np.array([[1, 1, 1, 1],
                         [1, 5, 1, 1],
                         [1, 4, 1, 5],
                         [1, 1, 1, 4]])
        d = Dendrogram.compute(data, min_value=2)
        assert len(d.trunk) == 2
        self.assert_valid_index(d, TreeIndex(d))

    def test_single_structure(self):
        data = np.array([[1, 1, 1, 1],
                         [1, 5, 1, 1],
                         [1, 4, 1, 1],
                         [1, 1, 1, 1]])
        d = Dendrogram.compute(data)
        assert len(d) == 1
        self.assert_valid_index(d, TreeIndex(d))

    def test_cube(self):
        np.random.seed(42)
        data = np.random.random((5, 5, 5))

        d = Dendrogram.compute(data)
        self.assert_valid_index(d, TreeIndex(d))

    def test_1d(self):
        np.random.seed(42)
        data = np.random.random(5)

        d = Dendrogram.compute(data)
        assert len(d) > 0
        self.assert_valid_index(d, TreeIndex(d))

    def test_4d(self):
        np.random.seed(42)
        data = np.random.random((3, 3, 3, 3))

        d = Dendrogram.compute(data)
        assert len(d) > 0
        self.assert_valid_index(d, TreeIndex(d))



# Licensed under an MIT open source license - see LICENSE

" Test import and export of dendrograms "

import os

import pytest
import numpy as np

from .. import Dendrogram
from ..structure import Structure
from .test_index import assert_permuted_fancyindex


class TestIO(object):

    def setup_method(self, method):
        n = np.nan
        self.data = np.array([[[n, n, n, n, n, n, n, n],
                               [n, 4, n, n, n, n, n, n],
                               [n, n, n, 1, n, n, 0, 5],
                               [3, n, n, 2, 3, 2, 0, n]],
                              [[n, n, n, n, n, n, n, n],
                               [1, n, n, n, n, n, n, n],
                               [1, n, 1, 1, 0, n, 0, 1],
                               [2, n, n, 1, 3, 1, n, 1]],
                              [[n, 2, 3, 4, n, n, n, n],
                               [1, 1, n, n, n, n, n, n],
                               [n, n, n, n, n, n, 0, 1],
                               [n, n, n, 1, 0, 1, 0, n]]])
        self.test_filename = None

    def teardown_method(self, method):
        if self.test_filename and os.path.exists(self.test_filename):
            os.remove(self.test_filename)

    def compare_dendrograms(self, d1, d2):
        " Helper method that ensures d1 and d2 are equivalent "
        # Do we get the same number of structures?
        assert len(d1) == len(d2)
        # Do we recover the data exactly?
        np.testing.assert_array_equal(d1.data, d2.data)
        # Now check that the structures are the same:
        for s in d2:
            idx = s.idx
            structure1, structure2 = d1[idx], d2[idx]
            assert_permuted_fancyindex(structure1.indices(subtree=False),
                                       structure2.indices(subtree=False))
            assert np.all(np.sort(structure1.values(subtree=False)) ==
                          np.sort(structure2.values(subtree=False)))
            assert isinstance(structure1, type(structure2))
            # Compare the coordinates and data values of all peak pixels:
            assert structure1.get_peak(subtree=True) == \
                structure2.get_peak(subtree=True)

            assert structure2._tree_index is not None

    # Below are the actual tests for each import/export format:

    def test_hdf5(self):
        self.test_filename = 'astrodendro-test.hdf5'
        d1 = Dendrogram.compute(self.data, verbose=False)
        d1.save_to(self.test_filename, format='hdf5')
        d2 = Dendrogram.load_from(self.test_filename, format='hdf5')
        self.compare_dendrograms(d1, d2)

    def test_fits(self):
        self.test_filename = 'astrodendro-test.fits'
        d1 = Dendrogram.compute(self.data, verbose=False)
        d1.save_to(self.test_filename, format='hdf5')
        d2 = Dendrogram.load_from(self.test_filename, format='hdf5')
        self.compare_dendrograms(d1, d2)

    def test_hdf5_auto(self):

        d1 = Dendrogram.compute(self.data, verbose=False)

        # recognize from extension
        d1.save_to('astrodendro-test.hdf5')

        # no way to tell
        with pytest.raises(IOError):
            d1.save_to('astrodendro-test')

        # no way to tell, so have to explicitly give format
        d1.save_to('astrodendro-test', format='hdf5')

        # recognize from extension
        d2 = Dendrogram.load_from('astrodendro-test.hdf5')

        # recognize from signature
        d2 = Dendrogram.load_from('astrodendro-test')

    def test_fits_auto(self):

        d1 = Dendrogram.compute(self.data, verbose=False)

        # recognize from extension
        d1.save_to('astrodendro-test.fits')

        # no way to tell
        with pytest.raises(IOError):
            d1.save_to('astrodendro-test')

        # no way to tell, so have to explicitly give format
        d1.save_to('astrodendro-test', format='fits')

        # recognize from extension
        d2 = Dendrogram.load_from('astrodendro-test.fits')

        # recognize from signature
        d2 = Dendrogram.load_from('astrodendro-test')
